from typing import Any, Dict, List

from datasets import Dataset
from rich.markup import escape
from rich.panel import Panel

from .base import Explorer


class DetectionExplorer(Explorer):
    """
    Explorer for detection tasks.
    """

    def __init__(self, dataset: Dataset) -> None:
        super().__init__(dataset)

    def get_panels(self, example: Dict[str, Any]) -> List[Panel]:
        panels = []
        # Prompt panel
        if example["label"] != "human":
            panels.append(
                Panel(
                    escape(example["prompt"]),
                    title="[red3]Prompt",
                    border_style="red3",
                    style="white",
                )
            )

        # Text panel
        panels.append(
            Panel(
                escape(example["text"]),
                title="[turquoise2]Text",
                border_style="turquoise2",
                style="white",
            )
        )
        return panels

    def get_title(self, idx: int, example: Dict[str, Any]) -> str:
        label, model = example["label"], example["model"]
        domain, extractor = example["domain"], example["extractor"]
        title = f"[tan]{idx}[/]: "
        if label == "generated":
            title += f"[violet]{label}[/] text ([light_coral]{model}[/]) "
        else:
            title += f"[violet]{label}[/] text "

        title += f"in domain [sea_green3]{domain}[/]"

        if label != "human":
            title += f" using extractor [yellow]{extractor}"

        return title



_MAJOR = "0"
_MINOR = "2"
_REVISION = "10"

VERSION_SHORT = "{0}.{1}".format(_MAJOR, _MINOR)
VERSION = "{0}.{1}.{2}".format(_MAJOR, _MINOR, _REVISION)

__version__ = VERSION


def _is_newer_than(version: str) -> bool:
    """True if current version is newer than 'version'."""
    from pkg_resources import parse_version as parse

    return parse(VERSION) > parse(version)


def _main():
    """
    For use inside an Azure pipeline.

    Usage:

    $ pip index versions text-machina | python -m text_machina.version

    Outputs:
     - 1 if code version is newer than latest version on pip.
     - 0 otherwise.
    """
    import sys

    for line in sys.stdin.readlines():
        line = line.strip()
        if line.startswith("LATEST"):
            latest_version = line.split()[-1]
            print(int(_is_newer_than(latest_version)))
            return
    else:
        raise ValueError("No LATEST line in stdin.")


if __name__ == "__main__":
    _main()



from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from datasets import Dataset

from ..common.logging import get_logger
from ..config import Config
from ..constrainers import get_length_constrainer
from ..data import PromptedDatasetBuilder
from ..models import get_model

_logger = get_logger(__name__)


class DatasetGenerator(ABC):
    """
    Base class for dataset generators.
    """

    def __init__(self, config: Config) -> None:
        self.config = config
        self.model = get_model(self.config.model)
        self.prompter = PromptedDatasetBuilder(self.config)

    def generate(self) -> Dataset:
        """
        Generates a labeled dataset based on the provided config.

        Returns:
            Dataset: the dataset
        """
        generations, kwargs = self._generate()
        dataset = self._pack(generations, **kwargs)
        dataset = self.add_config_info(dataset)
        return dataset

    def _generate(self) -> Tuple[List[str], Dict]:
        """
        Generates a dataset based on the provided config.

        Returns:
            Tuple[List[str], Dict]: a tuple of the generated texts and
                additional arguments to use for dataset packing.
        """
        # prepare inputs
        prompted_dataset = self.prompter.build()
        _logger.info(
            f"This is how one input looks like: {prompted_dataset.prompted_texts[0]}"
        )

        # instantiate length constrainer
        length_constrainer = get_length_constrainer(
            texts=prompted_dataset.human_texts,
            model_name=self.config.model.model_name,
            provider=self.config.model.provider,
        )

        # constrain generation config
        generation_config = length_constrainer.constrain(self.config.generation)

        _logger.info(
            f"Generating completions for with args:\n"
            f"Model: {self.config.model}.\n"
            f"Args: {generation_config}"
        )

        # run generator
        generations = self.model.generate_completions(
            prompts=prompted_dataset.prompted_texts,
            generation_config=generation_config,
        )

        return generations, {"prompted_dataset": prompted_dataset}

    @abstractmethod
    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        """
        Builds a dataset by packing the texts accordingly to the task.

        Args:
            generations (List[str]): the generated texts.
            **kwargs: additional keyword arguments.

        Returns:
            Dataset: the final labeled dataset.
        """
        ...

    def add_config_info(self, dataset: Dataset) -> Dataset:
        """
        Adds config information to the dataset.

        Args:
            dataset (Dataset): the dataset to add config information to.
        Returns:
            Dataset: the dataset with config information added.
        """
        dataset = dataset.add_column(
            "config_path",
            [str(self.config.path)] * len(dataset),
        )
        dataset = dataset.add_column(
            "language",
            [str(self.config.input.language)] * len(dataset),
        )
        return dataset



from math import ceil
from random import randint
from typing import Any, Dict, List, Tuple

from datasets import Dataset

from ..common.exceptions import ExtractorInvalidArgs
from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .utils import spacy_pipeline


class WordGap(Extractor):
    """
    Extractor that fills the prompt template with a boundary of
    two word spans (left-side and right-side of a sampled word),
    and with the number of words the LLM has to generate in
    between the boundary word spans.

    This extractor needs two template placeholders:
        - {n}: will be filled with the number of words to generate
               between the boundary words.
        - {boundaries}: will be filled with the boundary words separated
                        by the gap token and newlines.
                        E.g., "words1 ____ words2"

    This extractor allows to pass the following arguments in the
    `extractor_args` field from the config:
        - gap_token (str): gap token, e.g., "____"
        - max_percentage_boundaries (float): max percentage of
          boundaries to sample from a text. In a text of N words,
          there will be N-1 possible boundaries of two word spans.
        - max_word_span (int): max number of words to be generated
          between the boundary words.
        - range_boundary_size (List[float, float]): range where
          to sample the length of the word spans in the boundaries.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        args: Dict[str, Any] = input_config.extractor_args.get("word_gap", {})
        workspace: Dict[str, Any] = {
            "positions": [],
            "human_spans": [],
            "num_boundaries": [],
        }
        super().__init__(input_config, task_type, workspace, args)

    def check_valid_args(self):
        mandatory_args = [
            "gap_token",
            "max_percentage_boundaries",
            "max_word_span",
            "range_boundary_size",
        ]
        for mandatory_arg in mandatory_args:
            if mandatory_arg not in self.args:
                raise ExtractorInvalidArgs(
                    self.__class__.__name__, mandatory_args
                )

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        return [
            "".join(doc_words) for doc_words in self.workspace["human_spans"]
        ]

    def _format_boundary(self, pair: Tuple[str, str]) -> str:
        return f"{pair[0]} {self.args['gap_token']} {pair[1]}"

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        text_column = self.input_config.dataset_text_column
        texts = spacy_pipeline(
            texts=dataset[text_column], language=self.input_config.language
        )
        boundaries = []
        n_generated_words = []
        for text in texts:
            words = [word.text_with_ws for word in text]
            self.workspace["human_spans"].append(words)
            self.workspace["positions"].append([])
            # If the text has more than one word, it can be used
            # to interleave generations w/ human words.
            max_boundaries_to_select = ceil(
                (len(words) - 1) * self.args["max_percentage_boundaries"]
            )
            accum = 0
            if len(words) > 1:
                idx = 0
                while idx < len(words) - 1:
                    is_boundary = randint(0, 1)
                    boundary_size = 0
                    if is_boundary and accum < max_boundaries_to_select:
                        gen_sents = randint(1, self.args["max_word_span"])
                        boundary_size = randint(
                            *self.args["range_boundary_size"]
                        )
                        boundary = (
                            "".join(
                                words[max(0, idx - boundary_size) : idx + 1]
                            ),
                            "".join(words[idx + 1 : idx + 1 + boundary_size]),
                        )
                        boundaries.append(self._format_boundary(boundary))
                        self.workspace["positions"][-1].append(idx)
                        n_generated_words.append(gen_sents)
                        accum += 1
                    # Move far away to avoid coherence conflicts
                    # between boundary generations
                    idx += 1 + boundary_size

            # At this point:
            # (1) No boundaries have been selected:
            #     -> consider all the text as human
            # (2) >=1 boundary have been selected:
            #     -> the text could be interleaved (gen, human)
            # (3) The text has only one word:
            #     -> consider all the text as human
            self.workspace["num_boundaries"].append(accum)

        return {
            "n": list(map(str, n_generated_words)),
            "boundaries": boundaries,
        }



from enum import Enum
from typing import Final

import torch
from transformers import BitsAndBytesConfig

GENERATION_ERROR: Final = "<error>"

QUANTIZATION_CONFIGS: Final = {
    "int4": {"quantization_config": BitsAndBytesConfig(load_in_4bit=True)},
    "int4_bf16": {
        "quantization_config": BitsAndBytesConfig(
            load_in_4bit=True, bnb_4bit_compute_dtype=torch.float16
        )
    },
    "int4_nf4": {
        "quantization_config": BitsAndBytesConfig(
            load_in_4bit=True, bnb_4bit_quant_type="nf4"
        )
    },
    "int8": {"quantization_config": BitsAndBytesConfig(load_in_8bit=True)},
    "int4_nested": {
        "quantization_config": BitsAndBytesConfig(
            load_in_4bit=True, bnb_4bit_use_double_quant=True
        )
    },
    "fp16": {"torch_dtype": torch.float16},
    "none": {},
}


class CompletionType(str, Enum):
    CHAT: str = "CHAT"
    COMPLETION: str = "COMPLETION"



from abc import ABC, abstractmethod
from typing import Any, Dict


class Constrainer(ABC):
    """
    Base class for constrainers.

    A constrainer is any kind of class that infers something from a dataset
    and constrains the generation parameters according to that. For instance,
    length constrainers, that automatically infer the length and return maximum
    or minimum number of tokens accordingly.
    """

    def __init__(self):
        pass

    @abstractmethod
    def get_constraints(self) -> Dict[str, Any]:
        """
        Method that return parameters with values to constrain later the
        generation parameters.

        Example:
            output: {"max_tokens": 137, "min_tokens": 32}

        Returns:
            Dict[str, Any]: values to constrain generation parameters.
        """
        ...

    @abstractmethod
    def estimate(self) -> Any:
        """
        Method to estimate values that will be used in `constrain`.

        Returns:
            Any: any kind of value.
        """
        ...

    def constrain(self, generation_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Constrains a generation_config.

        Args:
            generation_config (Dict[str, Any]): a generation config.

        Returns:
            Dict[str, Any]: constrained generation config.
        """
        return {
            **self.get_constraints(),
            **generation_config,
        }



import atexit
import os
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from .cli_utils import _explore, _generate, generate_run_name, log_final_message
from .src.types import TaskType

os.environ["TOKENIZERS_PARALLELISM"] = "false"
app = typer.Typer()


@app.command()
def explore(
    config_path: Annotated[
        Path,
        typer.Option(
            help="The path to the YAML config or directory with YAML configs.",
            exists=True,
            file_okay=True,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
    task_type: Annotated[TaskType, typer.Option(help="The type of task.")],
    metrics_path: Annotated[
        Optional[Path],
        typer.Option(
            help="the path to the YAML metrics file.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ] = None,
    run_name: Annotated[
        Optional[str],
        typer.Option(
            help="The name of the run. Generated automatically if none given."
        ),
    ] = None,
    save_dir: Annotated[
        Optional[Path],
        typer.Option(help="The path where the exploration will be saved."),
    ] = None,
    max_generations: Annotated[
        int,
        typer.Option(
            help="The maximum number of texts to generate per config."
        ),
    ] = 10,
    interactive: Annotated[
        bool,
        typer.Option(
            help="Whether to step interactively through the generated texts"
        ),
    ] = True,
) -> None:
    """
    Generates a small set of texts, compares it against human texts,
    providing metrics and an interface for manual inspection of the generations
    """
    if not run_name:
        run_name = generate_run_name()
    if not save_dir:
        save_dir = Path.cwd() / run_name
    else:
        save_dir = save_dir / run_name

    atexit.register(log_final_message, run_name)

    _explore(
        config_path,
        metrics_path,
        save_dir,
        run_name,
        task_type,
        interactive,
        max_generations,
    )


@app.command()
def generate(
    config_path: Annotated[
        Path,
        typer.Option(
            help="The path to the YAML config or directory with YAML configs.",
            exists=True,
            file_okay=True,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
    task_type: Annotated[TaskType, typer.Option(help="The type of task.")],
    run_name: Annotated[
        Optional[str],
        typer.Option(
            help="The name of the run. Generated automatically if none given.",
        ),
    ] = None,
    save_dir: Annotated[
        Optional[Path],
        typer.Option(
            help=("The path where the generations will be saved."),
        ),
    ] = None,
) -> None:
    """
    Generates a dataset from the provided config or directory with configs.
    """
    if not run_name:
        run_name = generate_run_name()
    if not save_dir:
        save_dir = Path.cwd() / run_name
    else:
        save_dir = save_dir / run_name

    atexit.register(log_final_message, run_name)
    _generate(config_path, save_dir, run_name, task_type)


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import Dict, List

import pandas as pd
from datasets import Dataset

from ..common.exceptions import InvalidTaskTypeForMetric
from ..types import TaskType
from .base import Metric


class RepetitionDiversityMetric(Metric):
    """
    Implements the repetition and diversity metrics.
    See Sec. 4.1.2: https://arxiv.org/pdf/2202.06417.pdf

    Supported tasks: detection, attribution, and boundary.
    """

    def _run(self, dataset: Dataset, **kwargs) -> pd.DataFrame:
        # If we evaluate complete texts we can run easily grouping by labels
        if self.task_type in {TaskType.DETECTION, TaskType.ATTRIBUTION}:
            label = dataset["label"]
            text = dataset["text"]
        # Otherwise we need to separate the sections that pertain to different labels
        elif self.task_type == TaskType.BOUNDARY:
            df = dataset.to_pandas()[["text", "label"]]

            human_text = df.apply(
                lambda x: x["text"][: x["label"]], axis=1
            ).tolist()
            generated_text = df.apply(
                lambda x: x["text"][x["label"] :], axis=1
            ).tolist()

            text = human_text + generated_text
            label = ["human"] * len(human_text) + ["generated"] * len(
                generated_text
            )
        else:
            raise InvalidTaskTypeForMetric(self.name, self.task_type)

        ngrams = kwargs.get("ngrams", [2, 3, 4])

        results = []
        for text, label in zip(text, label):
            result = self.repetition_and_diversity(text, ngrams)
            result["label"] = label
            results.append(result)

        return pd.DataFrame(results)

    def _save(self, outputs: pd.DataFrame, path: Path) -> None:
        outputs.to_csv(path / "full_outputs.csv")
        outputs.drop("label", axis=1).mean(axis=0).to_json(
            path / "summary.json", indent=4
        )
        outputs.groupby("label").mean().reset_index().to_json(
            path / "per_label_summary.json", indent=4
        )

    def _log(self, outputs: pd.DataFrame, logger) -> None:
        summary = outputs.groupby("label").mean().reset_index()
        if self.task_type == TaskType.BOUNDARY:
            logger.info(
                f"Mean rep-n and diversity of segments: {summary.to_dict()}"
            )
        else:
            logger.info(f"Mean rep-n and diversity: {summary.to_dict()}")

    def repetition_and_diversity(self, text: str, ns: List[int]) -> Dict:
        tokens = text.strip().split()

        result = {}

        diversity = 1.0
        for n in ns:
            start_range = range(len(tokens) - n + 1)
            end_range = range(n, len(tokens) + 1)

            current_ngrams = [
                tuple(tokens[i:j]) for i, j in zip(start_range, end_range)
            ]

            uniques = len(set(current_ngrams))
            total = len(current_ngrams)

            if total > 0:
                ratio = uniques / total
            else:
                ratio = 0

            diversity *= ratio
            result[f"rep-{n}"] = 100 * (1.0 - (ratio))

        result["div"] = diversity

        return result



from typing import List

from transformers import AutoTokenizer

from .base import Tokenizer


class HuggingFaceLocalTokenizer(Tokenizer):
    """
    Tokenizer for HuggingFace models.
    """

    def __init__(self, model_name: str):
        super().__init__(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)

    def encode(self, text: str) -> List[int]:
        return self.tokenizer.encode(text)

    def decode(self, tokens: List[int]) -> str:
        return self.tokenizer.decode(tokens, skip_special_tokens=True)



from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import numpy as np

from .base import Constrainer


class LengthConstrainer(Constrainer, ABC):
    """
    Base class for length constrainers to compute
    automatically the number of tokens to generate.
    """

    def __init__(
        self,
        lengths: List[int],
        provider: str,
        min_tokens: int = 10,
        max_tokens: int = 512,
    ):
        super().__init__()
        self.provider = provider
        self.lengths = lengths
        self.min_tokens = min_tokens
        self.max_tokens = max_tokens

    @abstractmethod
    def estimate(self) -> Tuple[int, int]:
        """
        Estimates min and max token lengths.

        Returns:
            Tuple[int, int]: tuple with the min and max token lengths.
        """
        ...

    def get_constraints(self) -> Dict[str, int]:
        """
        Obtains min and max token lengths, and returns
        these params according to the provider.

        Returns:
            Dict[str, int]: dict with the token length params.
        """
        min_new_tokens, max_new_tokens = self.estimate()

        if self.provider in [
            "openai",
            "azure_openai",
            "anthropic",
            "cohere",
            "ai21",
            "inference_server",
        ]:
            return {"max_tokens": max_new_tokens}
        elif self.provider == "vertex":
            return {"max_output_tokens": max_new_tokens}
        elif self.provider == "bedrock":
            return {"maxTokenCount": max_new_tokens}
        return {
            "max_new_tokens": max_new_tokens,
            "min_new_tokens": min_new_tokens,
        }


class MeanLengthConstrainer(LengthConstrainer):
    """
    Constrainer within one std radius of the mean of a list of numbers
    """

    def __init__(
        self,
        lengths: List[int],
        provider: str,
        min_tokens: int = 10,
        max_tokens: int = 512,
    ):
        super().__init__(lengths, provider, min_tokens, max_tokens)

    def estimate(self) -> Tuple[int, int]:
        mean = np.mean(self.lengths)
        std = np.std(self.lengths)

        max_new_tokens = int(min(self.max_tokens, mean + 2 * std))
        min_new_tokens = int(max(self.min_tokens, mean - 2 * std))
        return (min_new_tokens, max_new_tokens)


class MedianLengthConstrainer(LengthConstrainer):
    """
    Constrainer within one std radius of the median of a list of numbers
    """

    def __init__(
        self,
        lengths: List[int],
        provider: str,
        min_tokens: int = 10,
        max_tokens: int = 512,
    ):
        super().__init__(lengths, provider, min_tokens, max_tokens)

    def estimate(self) -> Tuple[int, int]:
        median = np.median(self.lengths)
        std = np.std(self.lengths)

        max_new_tokens = int(min(self.max_tokens, median + 2 * std))
        min_new_tokens = int(max(self.min_tokens, median - 2 * std))
        return (min_new_tokens, max_new_tokens)



from abc import ABC, abstractmethod
from typing import Any, Dict, List

from datasets import Dataset
from readchar import key, readkey
from rich import print as rprint
from rich.console import Group
from rich.panel import Panel


class Explorer(ABC):
    """
    Base class for explorers that run an interactive
    CLI to visualize generated datasets.
    """

    def __init__(self, dataset: Dataset) -> None:
        self.dataset = dataset

    @property
    def command_panel(self) -> Panel:
        """
        Command panel to help the users with commands.
        Returns:
            Panel: panel showing the commands and action keys.
        """
        return Panel(
            "Prev: [yellow3]left arrow[/], [yellow3]A[/], or [yellow3]backspace[/].\n"
            "Next: [yellow3]right arrow[/], [yellow3]D[/], or [yellow3]enter[/].\n"
            "Quit: [yellow3]q[/].",
            expand=False,
        )

    def step(self) -> None:
        """
        Steps through a generated dataset, handling user input and
        showing the texts in a command line UI.
        """
        idx = 0
        N = len(self.dataset)

        self.show_example(idx)

        while idx < len(self.dataset):
            idx = self.get_next_position_or_exit(idx, max_position=N - 1)
            self.show_example(idx)

    def get_next_position_or_exit(
        self, position: int, max_position: int
    ) -> int:
        """
        Handles user input to cycle through a dataset.

        Args:
            current_position (int): the current position.
        Returns:
            int: the new position.
        """
        user_input = readkey()

        if user_input in {key.LEFT, "a", "A", key.BACKSPACE, key.DELETE}:
            position = max(position - 1, 0)
        elif user_input in {key.RIGHT, "d", "D", key.ENTER, key.CR, key.TAB}:
            position = min(position + 1, max_position)
        elif user_input in {"q", "Q"}:
            exit()

        return position

    def show_example(self, idx: int) -> None:
        """
        Pretty prints an example from the dataset.

        Args:
            idx: index of the example in the dataset.
        """
        print("\033[H\033[J", end="")
        panels = self.get_panels(self.dataset[idx])
        title = self.get_title(idx, self.dataset[idx])
        rprint(Panel(Group(*panels), title=title))
        rprint(self.command_panel)

    @abstractmethod
    def get_panels(self, example: Dict[str, Any]) -> List[Panel]:
        """
        Builds the list of panels for the elements (prompts, texts, etc.)
        to be printed. This method should be implemented by each
        task-specific explorer.

        Args:
            example (Dict[str, Any]): an example from the dataset.
        Returns:
            List[Panel]: lists of panels to be printed in the desired order.
        """
        ...

    @abstractmethod
    def get_title(self, idx: int, example: Dict[str, Any]) -> str:
        """
        Builds a printable title based on an example from the dataset.
        This method should be implemented by each task-specific explorer.

        Args:
            idx: index of the example.
            example (Dict[str, Any]): an example from the dataset.
        Returns:
            str: a printable title.
        """
        ...



# flake8: noqa
from typing import Mapping, Type

from datasets import Dataset

from ..types import TaskType
from .attribution import AttributionExplorer
from .base import Explorer
from .boundary import BoundaryExplorer
from .detection import DetectionExplorer
from .mixcase import MixCaseExplorer

EXPLORERS: Mapping[str, Type[Explorer]] = {
    TaskType.DETECTION: DetectionExplorer,
    TaskType.ATTRIBUTION: AttributionExplorer,
    TaskType.BOUNDARY: BoundaryExplorer,
    TaskType.MIXCASE: MixCaseExplorer,
}


def get_explorer(task_type: TaskType, dataset: Dataset) -> Explorer:
    """
    Gets an explorer from the pool.

    Args:
        task_type (TaskType): the task type.
        dataset (Dataset): a dataset.

    Returns:
        Explorer: an explorer from the pool.
    """
    return EXPLORERS[task_type](dataset)


__all__ = [str(cls) for cls in EXPLORERS.values()]



from random import randint
from typing import Dict, List

from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .utils import spacy_pipeline


class SentencePrefix(Extractor):
    """
    Extractor that fills the prompt template with a sentence
    prefix extracted from a text column of a dataset.

    This extractor needs a template placeholder named {sentences}.

    This extractor allows to pass the following arguments in the
    `extractor_args` field from the config:
        - k (int): number of sentences in the prefix. If not specified,
                   `k` will be random for each sample.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)
        self.args = self.input_config.extractor_args.get("sentence_prefix", {})
        self.n_sentences = lambda x: self.args.get(
            "k", randint(1, max(1, len(x) - 1))
        )
        self.sampled_positions: List[int] = []

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        """
        For detection and attribution tasks, removes the extracted prefix
        from human texts to ensure both generations and human texts are
        continuations of sentence prefixes.

        For boundary tasks (human followed by generated), returns the prefix.

        Args:
            human_texts (List[str]): list of human texts.

        Returns:
            List[str]: prepared human texts.
        """
        docs = spacy_pipeline(
            human_texts,
            language=self.input_config.language,
            disable_pipes=[
                "ner",
                "tagger",
                "attribute_ruler",
                "lemmatizer",
            ],
        )
        output: List[str] = []
        for idx, doc in enumerate(docs):
            doc_sents = list(doc.sents)
            n_sentences = self.sampled_positions[idx]
            if self.task_type in [TaskType.DETECTION, TaskType.ATTRIBUTION]:
                text = "".join(
                    sent.text_with_ws for sent in doc_sents[n_sentences:]
                )
            else:
                text = "".join(
                    sent.text_with_ws for sent in doc_sents[:n_sentences]
                )
            output.append(text)
        return output

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        docs = spacy_pipeline(
            dataset[self.input_config.dataset_text_column],
            language=self.input_config.language,
            disable_pipes=[
                "ner",
                "tagger",
                "attribute_ruler",
                "lemmatizer",
            ],
        )
        output_texts = []
        for doc in docs:
            doc_sents = list(doc.sents)
            n_sentences = self.n_sentences(doc_sents)
            self.sampled_positions.append(n_sentences)
            output_texts.append(
                "".join([sent.text_with_ws for sent in doc_sents[:n_sentences]])
            )

        return {"sentences": output_texts}



import hashlib
import re
from functools import wraps
from typing import Callable, Dict, List

import fasttext
import ftfy
import numpy as np
import pandas as pd
import requests
from datasets import Dataset, disable_caching

from .common.logging import get_logger
from .common.utils import get_cache_path
from .models.types import GENERATION_ERROR
from .types import TaskType

_logger = get_logger(__name__)

disable_caching()

# Monkeypatch to disable fasttext warning:
# https://stackoverflow.com/questions/66353366/cant-suppress-fasttext-warning-load-model-does-not-return
fasttext.FastText.eprint = lambda x: None


def batched_map(
    f: Callable[[List[str]], Dict[str, List[str]]]
) -> Callable[[Dataset], Dataset]:
    """
    Runs a function `f` on a dataset with batched mapping.

    Args:
        f (Callable[[List[str]], Dict[str, List[str]]]): the function.

    Returns:
        Callable[[Dataset], Dataset]: the modified function.
    """

    @wraps(f)
    def with_batched_mapping(dataset: Dataset) -> Dataset:
        """Applies batched map to a function, shows name in progress bar"""
        desc = " ".join(f.__name__.split("_")).capitalize()
        dataset = dataset.map(
            f,
            input_columns=["text"],
            batched=True,
            load_from_cache_file=False,
            desc=desc,
        )
        return dataset

    return with_batched_mapping


def remove_generation_errors(dataset: Dataset) -> Dataset:
    """
    Removes generation errors, i.e. texts marked with `GENERATION_ERROR`.

    Args:
        dataset (Dataset): the dataset to filter.
    Returns:
        Dataset: a filtered dataset with no error annotations.
    """
    return dataset.filter(lambda x: x["text"] != GENERATION_ERROR)


def truncate(
    dataset: Dataset,
    min_length: int = 5,
    min_tokens_to_truncate: int = 2,
    sampling_radius_size: float = 2.0,
) -> Dataset:
    """
    Truncates texts to remove token length bias per class in each domain.

    This is done by:
    1. Sampling the same number of texts per label in each domain
    2. Sorting them by token length
    3. Grouping them such that each group has one text per label
    4. Truncating the texts in the group to have the same length
    5. Truncating the remainder of 1. between mean +- 2*std
    6. Dropping texts with lengths < min_length

    Note that all the texts are truncated, this can be modified with
        min_tokens_to_truncate = 0.

    Args:
        dataset (Dataset): the dataset to truncate.
        min_length (int): the minimum (spacy) token length.
        min_tokens_to_truncate (int): the minimum (spacy) tokens to truncate.
        sampling_radius_size (float): the radius size for the sampling of
            token lengths for non-grouped texts.

    Returns:
        Dataset: the truncated dataset
    """
    from .extractors.utils import spacy_pipeline

    np.random.seed(0)

    df = dataset.to_pandas()

    # tokenize with spacy multilingual model
    df["tokenized"] = spacy_pipeline(
        df["text"],
        "multilingual",
    )
    # get tokenized length: we'll keep track of this and a
    # "difference" column to know how much to truncate
    # so then we only have to run the truncation itself once
    df["token_length"] = df["tokenized"].apply(len)

    to_truncate = []
    # set new token lengths (actual truncation happens later)
    for domain in df["domain"].unique():
        # sample enough data per label
        min_size = df[df["domain"] == domain].groupby("label").size().min()
        grouped = df[df["domain"] == domain].groupby("label").sample(min_size)

        # the remainder will be truncated based on mean +-std estimations
        remainder = df[~df.index.isin(grouped.index)].copy()
        remainder = remainder[remainder["domain"] == "domain"]

        # texts with similar token lengths are grouped together
        # to better approximate domain token length distribution
        # and truncate less from longer texts.
        grouped["group"] = grouped.groupby("label")["token_length"].rank(
            "first"
        )

        # truncate a text by setting its new length to:
        # min_token_length_per_group - min_tokens_to_truncate
        grouped["new_token_length"] = (
            grouped.groupby("group")["token_length"].transform(min)
            - min_tokens_to_truncate
        )
        grouped = grouped.drop(["group"], axis=1)

        # compute mean and std to tuncate the ungrouped texts
        # for this we must only consider new lengths within desired min and max
        new_lengths_above_min = grouped["new_token_length"][
            (grouped["new_token_length"] >= min_length)
        ]
        mean, std = new_lengths_above_min.mean(), new_lengths_above_min.std()

        # Sample lengths around mean +- 2 std
        low = int(max(min_length, mean - sampling_radius_size * std))
        high = int(mean + sampling_radius_size * std)
        sampled_new_token_lengths = np.random.randint(
            low=low, high=high, size=len(remainder)
        )

        # account for cases where the sampled token length > current token length
        remainder["new_token_length"] = np.minimum(
            sampled_new_token_lengths,
            remainder["token_length"] - min_tokens_to_truncate,
        )

        both = pd.concat([grouped, remainder])

        # drop rows not within length bounds
        both = both[(both["new_token_length"] >= min_length)]

        to_truncate.append(both)

    new_df = pd.concat(to_truncate).reset_index(drop=True)

    assert (new_df["token_length"] < new_df["new_token_length"]).sum() == 0

    def truncate_and_decode_one(row: Dict) -> str:
        tokenized_truncated = row["tokenized"][: row["new_token_length"]]
        text = "".join([token.text_with_ws for token in tokenized_truncated])
        return text

    new_df["text"] = new_df.apply(truncate_and_decode_one, axis=1)

    diff = new_df["token_length"] - new_df["new_token_length"]
    dropped_quantity = len(df) - len(new_df)

    _logger.info(
        f"Truncated texts. Length difference statistics: {diff.describe().to_dict()}"
    )
    if dropped_quantity:
        _logger.info(
            f"{dropped_quantity} texts were too short and were dropped in the truncation process."
        )

    new_df = new_df.drop(
        ["token_length", "new_token_length", "tokenized"], axis=1
    )

    dataset = Dataset.from_pandas(new_df)
    return dataset


def filter_by_language(dataset: Dataset, language: str = "en") -> Dataset:
    """
    Applies a language id filter, removing texts in undesired languages.

    Args:
        dataset (Dataset): the dataset to apply the language id filter on.
    Returns:
        Dataset: the filtered dataset.
    """
    model = get_langid_model()

    @batched_map
    def annotate_texts_with_language(texts: List[str]):
        keep = []
        for text in texts:
            # need to replace newlines since fasttexts doesn't accept them
            predicted_language = model.predict(text.replace("\n", " "), k=1)

            # fasttext model output looks like: (('__label__en',), array([0.98803425]))
            # so we grab the suffix with the language ISO code
            predicted_language = predicted_language[0][0][-2:]
            if predicted_language != language:
                keep.append(False)
            else:
                keep.append(True)

        return {"keep": keep}

    dataset = annotate_texts_with_language(dataset)
    dataset = dataset.filter(lambda x: x["keep"], desc="Filter by language")
    dataset = dataset.remove_columns(["keep"])

    return dataset


def get_langid_model() -> fasttext.FastText:
    url = (
        "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"
    )
    model_path = get_cache_path() / "fasttext" / "lid.176.bin"

    # Need to download if it doesn't exist
    try:
        model = fasttext.load_model(str(model_path))
    except (ValueError, FileNotFoundError):
        model_path.parent.mkdir(parents=True, exist_ok=True)
        content = requests.get(url, stream=True).content
        with model_path.open("wb") as f:
            f.write(content)

        model = fasttext.load_model(str(model_path))

    return model


def remove_label_duplicates(dataset: Dataset) -> Dataset:
    """
    Removes text with more than one associated label.

    Args
        dataset (Dataset): the dataset to remove label duplicates from.

    Returns:
        Dataset: the dataset with removed label duplicates.
    """
    old_len = len(dataset)
    df = dataset.to_pandas()

    groups = df.groupby("text")["label"].nunique()
    single_label_texts = groups[groups == 1].index
    filtered = df[df["text"].isin(single_label_texts)]

    dataset = Dataset.from_pandas(filtered, preserve_index=False)

    new_len = len(dataset)
    _logger.info(f"Removed {old_len - new_len} texts with more than one label.")

    return dataset


def remove_text_duplicates(dataset: Dataset) -> Dataset:
    """
    Removes all text duplicates.

    Args
        dataset (Dataset): the dataset to remove duplicates from.

    Returns:
        Dataset: the dataset with removed duplicates.
    """
    old_len = len(dataset)

    def hash_str(s: str) -> str:
        x = s.encode("utf8")
        return hashlib.sha256(x).hexdigest()

    dataset = dataset.map(
        lambda example: {"hash": hash_str(example["text"])},
        load_from_cache_file=False,
    )

    _, unique_indices = np.unique(dataset["hash"], return_index=True)

    dataset = dataset.select(unique_indices)

    dataset = dataset.remove_columns(["hash"])
    new_len = len(dataset)
    _logger.info(f"Removed {old_len - new_len} duplicated texts.")

    return dataset


def remove_empty_texts(dataset: Dataset) -> Dataset:
    """
    Removes empty texts from a dataset.

    Args:
        dataset (Dataset): the dataset to remove empty texts from.

    Returns:
        Datset: the dataset with removed empty texts.
    """
    old_len = len(dataset)
    dataset = dataset.filter(
        lambda example: example["text"], load_from_cache_file=False
    )
    new_len = len(dataset)
    _logger.info(f"Removed {old_len - new_len} empty texts.")
    return dataset


@batched_map
def remove_special_tokens(texts: List[str]) -> Dict[str, List[str]]:
    """
    Removes special text generation tokens from a list of texts.

    Args:
        texts (List[str]): the texts to apply special-token removal to.
    Returns:
        Dict[str, List[str]]: the cleaned texts in dict form.
        The result is returned as so in order to run this using
        batched mapping from huggingface datasets.
    """
    special_tokens = [
        "[CLS]",
        "[SEP]",
        "[PAD]",
        "[MASK]",
        "[UNK]",
        "[BOS]",
        "[EOS]",
        "[EOD]",
        "[EOP]",
        "<endoftext>",
    ]
    # Also remove brackets from ends
    special_tokens += [x[1:-1] for x in special_tokens]

    regex = re.compile("|".join(map(re.escape, special_tokens)))
    clean = []
    for text in texts:
        text = regex.sub("", text)
        clean.append(text)

    return {"text": clean}


@batched_map
def remove_disclosure_phrases(texts: List[str]) -> Dict[str, List[str]]:
    """
    Removes a set of disclosure phrases.

    Args:
        texts (List[str]): the texts from which to remove disclosure phrases.
    Returns:
        Dict[str, List[str]]: the cleaned texts in dict form.
        The result is returned as so in order to run this using
        batched mapping from huggingface datasets.
    """

    # These patterns match the prefixes until next word
    patterns = [
        r"As an AI language model.+?(?=\w)",
        r"I am sorry, I am an AI language model and.+?(?=\w)",
        r"I am sorry, I'm an AI language model and.+?(?=\w)",
        r"I'm sorry, I'm an AI language model and.+?(?=\w)",
        r"I'm sorry, I am an AI language model and.+?(?=\w)",
        r"I'm sorry, but I am an AI language model and.+?(?=\w)",
    ]
    regexes = [re.compile(pattern, flags=re.IGNORECASE) for pattern in patterns]

    clean = []
    for text in texts:
        original_text = text

        for regex in regexes:
            text = regex.sub("", text)

        # At least one regex was applied so we correct the capitalization
        if original_text != text:
            text = text.capitalize()

        clean.append(text)

    return {"text": clean}


@batched_map
def fix_encoding(texts: List[str]) -> Dict[str, List[str]]:
    """
    Fixes the encoding in a list of texts.

    Args:
        texts (List[str]): the texts to apply encoding-fixing to.
    Returns:
        Dict[str, List[str]]: the cleaned texts in dict form.
        The result is returned as so in order to run this using
        batched mapping from huggingface datasets.
    """
    clean = []
    for text in texts:
        text = ftfy.fix_text(text)
        clean.append(text)

    return {"text": clean}


@batched_map
def strip(texts: List[str]) -> Dict[str, List[str]]:
    """
    Strips whitespace from a list of texts.

    Args:
        texts (List[str]): the texts to apply stripping to.
    Returns:
        Dict[str, List[str]]: the cleaned texts in dict form.
        The result is returned as so in order to run this using
        batched mapping from huggingface datasets.
    """
    clean = []
    for text in texts:
        text = text.strip()
        clean.append(text)
    return {"text": clean}


def postprocess(dataset: Dataset, task_type: TaskType) -> Dataset:
    """
    Postprocesses a dataset.

    Args:
        dataset (Dataset): the dataset to postprocess.
    Returns:
        Dataset: the postprocessed dataset.
    """

    single_text_actions = [
        fix_encoding,
        strip,
        remove_special_tokens,
        remove_disclosure_phrases,
    ]

    full_dataset_actions = [
        remove_generation_errors,
        remove_empty_texts,
        remove_text_duplicates,
    ]

    if task_type in {TaskType.DETECTION, TaskType.ATTRIBUTION}:
        if len(set(dataset["label"])) < 2:
            _logger.info(
                "Dataset only has single label, truncation will not be applied."
            )
            actions = (
                single_text_actions
                + full_dataset_actions
                + [remove_label_duplicates]
            )
        else:
            actions = (
                single_text_actions
                + [truncate]
                + full_dataset_actions
                + [remove_label_duplicates]
            )
    else:
        actions = single_text_actions + full_dataset_actions  # type: ignore

    for action in actions:
        dataset = action(dataset)  # type: ignore

    return dataset



import os

from openai import AzureOpenAI

from ..common.utils import get_instantiation_args
from ..config import ModelConfig
from .openai import OpenAIModel


class AzureOpenAIModel(OpenAIModel):
    """
    Generates completions using Azure OpenAI models.

    Requires the definition of the `AZURE_OPENAI_API_KEY=<key>` env variable.
    """

    def __init__(self, model_config: ModelConfig):
        self.model_config = model_config
        self.model_name = model_config.model_name
        self.client = AzureOpenAI(
            api_key=os.environ["AZURE_OPENAI_API_KEY"],
            **get_instantiation_args(
                AzureOpenAI.__init__, self.model_config.model_dump()
            ),
        )



from math import ceil
from random import randint, sample, uniform
from typing import Any, Dict, List

from datasets import Dataset

from ..common.exceptions import ExtractorInvalidArgs
from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .utils import spacy_pipeline


class SentenceRewriting(Extractor):
    """
    Extractor that fills the prompt template with a sentence that
    has to be rewritten by an LLM.

    This extractor needs two template placeholders:


    This extractor needs two template placeholders:
        - {sentence}: will be filled with sentence to be rewritten.

    This extractor allows to pass the following arguments in the
    `extractor_args` field from the config:
        - percentage_range (List[float]): range delimiting the percentage
            of sentences to be rewritten. At least one sentence will be
            always rewritten.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        args: Dict[str, Any] = input_config.extractor_args.get(
            "sentence_rewriting", {}
        )
        workspace: Dict[str, Any] = {"positions": [], "human_spans": []}
        super().__init__(input_config, task_type, workspace, args)

    def check_valid_args(self):
        mandatory_args = ["percentage_range"]
        for mandatory_arg in mandatory_args:
            if mandatory_arg not in self.args:
                raise ExtractorInvalidArgs(
                    self.__class__.__name__, mandatory_args
                )

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        return [
            "".join(doc_sentences)
            for doc_sentences in self.workspace["human_spans"]
        ]

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        text_column = self.input_config.dataset_text_column
        texts = spacy_pipeline(
            texts=dataset[text_column], language=self.input_config.language
        )
        sentences_to_rewrite = []
        for text in texts:
            sentences = list([sent.text_with_ws for sent in text.sents])
            # Skip the sample if there are no sentences.
            if len(sentences) == 0:
                continue
            self.workspace["human_spans"].append(sentences)
            n_sentences_to_select = randint(
                1,
                ceil(uniform(*self.args["percentage_range"]) * len(sentences)),
            )
            sampled_positions = sorted(
                sample(list(range(len(sentences))), n_sentences_to_select)
            )
            self.workspace["positions"].append(sampled_positions)
            for position in sampled_positions:
                sentences_to_rewrite.append(sentences[position])
        return {
            "sentence": list(map(str, sentences_to_rewrite)),
        }



from pathlib import Path
from typing import Dict, List

import evaluate
import numpy as np
import pandas as pd
from datasets import Dataset, DatasetDict
from transformers import (
    AutoModelForTokenClassification,
    AutoTokenizer,
    DataCollatorForTokenClassification,
    Trainer,
    TrainingArguments,
)

from ..common.exceptions import InvalidTaskTypeForMetric
from ..types import DetectionLabels, TaskType
from .base import Metric


def prepare_tags_for_mixcase(
    offset_mappings: List[List[List[int]]],
    labels: List[List[Dict]],
    label_mapping: Dict[str, int],
) -> Dict[str, List[List[int]]]:
    """
    Prepares the labels for mixcase tasks to be addressed
    as token classification tasks.

    This fn is designed to work with the `map` HF's function
    using `batched=True`.

    Args:
        offset_mappings (List[List[List[int]]]): offset mappings of each text.
        labels (List[List[Dict]]): labels of each text.
        label_mapping (Dict[str, int]): label mapping str to int labels.

    Returns:
        Dict[str, List[List[int]]]: the tags for each text in the batch.
    """
    tags = []
    for offset_mapping, label in zip(offset_mappings, labels):
        sample_tags = [-100]
        for _, char_end in offset_mapping[1:-1]:
            if char_end < label[0]["end"]:
                sample_tags.append(label_mapping[label[0]["label"]])
            else:
                if len(label) > 1:
                    label.pop(0)
                sample_tags.append(label_mapping[label[0]["label"]])
        sample_tags.append(-100)
        tags.append(sample_tags)
    return {"labels": tags}


def prepare_tags_for_boundary(
    offset_mappings: List[List[List[int]]],
    labels: List[int],
    label_mapping: Dict[str, int],
) -> Dict[str, List[List[int]]]:
    """
    Prepares the labels for boundary tasks to be addressed
    as token classification tasks.

    This fn is designed to work with the `map` HF's function
    using `batched=True`.

    Args:
        offset_mappings (List[List[List[int]]]): offset mappings of each text.
        labels (List[int): labels of each text.
        label_mapping (Dict[str, int]): label mapping str to int labels.

    Returns:
        Dict[str, List[List[int]]]: the tags for each text in the batch.
    """
    tags = []
    for offset_mapping, label in zip(offset_mappings, labels):
        sample_tags = [-100]
        for _, char_end in offset_mapping[1:-1]:
            if char_end <= label:
                sample_tags.append(label_mapping[DetectionLabels.HUMAN.value])
            else:
                sample_tags.append(
                    label_mapping[DetectionLabels.GENERATED.value]
                )
        sample_tags.append(-100)
        tags.append(sample_tags)
    return {"labels": tags}


def prepare_dataset(
    dataset: Dataset,
    task_type: TaskType,
    tokenizer: AutoTokenizer,
    test_size: float,
    label_mapping: Dict[str, int],
) -> DatasetDict:
    """
    Prepares the dataset (tokenization, prepare labels, splitting, etc.).

    Args:
        dataset (Dataset): a dataset.
        tokenizer (AutoTokenizer): a tokenizer.
        test_size (float): proportion reserved for the test set.
        label_mapping (Dict[str, int]): label mapping str to int labels.

    Returns:
        DatasetDict: a dataset with train and test splits.
    """
    dataset = dataset.map(
        lambda batch: tokenizer(
            batch, truncation=True, return_offsets_mapping=True
        ),
        input_columns=["text"],
        batched=True,
    )
    dataset = dataset.map(
        (
            prepare_tags_for_mixcase
            if task_type == TaskType.MIXCASE
            else prepare_tags_for_boundary
        ),
        input_columns=["offset_mapping", "label"],
        batched=True,
        fn_kwargs={"label_mapping": label_mapping},
    )
    dataset = dataset.select_columns(["input_ids", "attention_mask", "labels"])
    dataset = dataset.train_test_split(test_size=test_size)
    return dataset


def fit(
    model: AutoModelForTokenClassification,
    dataset: Dataset,
    tokenizer: AutoTokenizer,
    training_args: Dict,
) -> None:
    """
    Fits a model on a dataset.

    Args:
        model (AutoModelForTokenClassification): a model.
        dataset (Dataset): a training dataset.
        tokenizer (AutoTokenizer): a tokenizer.
        training_args (Dict): args to be passed to the HF's Trainer.
    """
    training_args = TrainingArguments(do_train=True, **training_args)
    collator = DataCollatorForTokenClassification(tokenizer=tokenizer)
    trainer = Trainer(
        model,
        train_dataset=dataset,
        data_collator=collator,
        args=training_args,
    )
    trainer.train()


def predict(
    model: AutoModelForTokenClassification,
    dataset: Dataset,
    tokenizer: AutoTokenizer,
) -> List[List[int]]:
    """
    Predicts a dataset using a model.

    Args:
        model (AutoModelForTokenClassification): a model.
        dataset (Dataset): a test dataset.
        tokenizer (AutoTokenizer): a tokenizer.

    Returns:
        List[List[int]]: list of predicted labels for each example.
    """
    collator = DataCollatorForTokenClassification(tokenizer=tokenizer)
    trainer = Trainer(
        model,
        data_collator=collator,
    )
    predictions = trainer.predict(dataset).predictions
    return np.argmax(predictions, axis=-1)


def eval(
    predictions: List[List[int]],
    references: List[List[int]],
    label_mapping: Dict[int, str],
) -> Dict[str, float]:
    """
    Evaluates using `seqeval` from HF metrics.

    Args:
        predictions (List[List[int]]): list of predictions for each example.
        references (List[List[int]]): list of gold labels for each example.
        label_mapping (Dict[int, str]): label mapping int to str labels.

    Returns:
        Dict[str, float] -> dictionary with metric values.
    """
    preds = [
        [label_mapping[p] for (p, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, references)
    ]

    refs = [
        [label_mapping[l] for (_, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(preds, references)
    ]

    preds = [pred[1:] for pred in preds]

    seqeval = evaluate.load("seqeval")
    results = seqeval.compute(predictions=preds, references=refs)

    return {
        "precision": results["overall_precision"],
        "recall": results["overall_recall"],
        "f1": results["overall_f1"],
        "accuracy": results["overall_accuracy"],
    }


class TokenClassificationMetric(Metric):
    """
    Implements a HF token classification model
    for evaluating a mixcase dataset.

    Supported tasks: boundary and mixcase.
    """

    def _run(self, dataset: Dataset, **kwargs) -> pd.DataFrame:
        if self.task_type not in [TaskType.MIXCASE, TaskType.BOUNDARY]:
            raise InvalidTaskTypeForMetric(self.name, self.task_type)

        model = AutoModelForTokenClassification.from_pretrained(
            **kwargs["model_args"]
        )
        tokenizer = AutoTokenizer.from_pretrained(
            kwargs["model_args"]["pretrained_model_name_or_path"]
        )

        dataset = prepare_dataset(
            dataset,
            self.task_type,
            tokenizer,
            kwargs["test_size"],
            kwargs["label_mapping"],
        )

        fit(model, dataset["train"], tokenizer, kwargs["training_args"])
        predictions = predict(model, dataset["test"], tokenizer)

        results = eval(
            predictions,
            dataset["test"]["labels"],
            {v: k for k, v in kwargs["label_mapping"].items()},
        )
        return pd.DataFrame(results, index=[0])

    def _save(self, outputs: pd.DataFrame, path: Path) -> None:
        outputs.to_csv(path / "full_outputs.csv", index=False)

    def _log(self, outputs: pd.DataFrame, logger) -> None:
        logger.info(f"Seqeval results: {outputs}")



from math import ceil
from random import choice, randint, uniform
from typing import Any, Dict, List

from datasets import Dataset

from ..common import color_log, get_logger
from ..common.exceptions import ExtractorInvalidArgs
from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .utils import spacy_pipeline

_logger = get_logger(__name__)


class WordMasking(Extractor):
    """
    Extractor that fills the prompt template with a text with
    masked word spans and the LLM has to generate all the masked
    word spans.

    This extractor needs two template placeholders:
        - {masked_text}: will be filled with a text with masked word spans.

    This extractor allows to pass the following arguments in the
    `extractor_args` field from the config:
        - mask_token (str): mask token, e.g., "MASK". Several masks in a text
            will be appended with the index, e.g. "MASK-0"
        - percentage_range (List[float]): range delimiting the percentage
            of word spans to be masked. At least one word span will be
            always masked.
        - span_length_range (List[int]): range where to sample the length
            of each masked span.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        args: Dict[str, Any] = input_config.extractor_args.get(
            "word_masking", {}
        )
        workspace: Dict[str, Any] = {"masked_texts": []}
        super().__init__(input_config, task_type, workspace, args)

        _logger.warn(
            color_log(
                f"You are using the `{self.__class__.__name__}` extractor."
                " Consider that few models like GPT-4 can work properly with this"
                " type of generation. Models must be:\n"
                "1) Capable of generating proper JSON.\n"
                "2) Capable enough to generate all the masks.",
                "bold_yellow",
            )
        )

    def check_valid_args(self):
        mandatory_args = ["mask_token", "percentage_range", "span_length_range"]
        for mandatory_arg in mandatory_args:
            if mandatory_arg not in self.args:
                raise ExtractorInvalidArgs(
                    self.__class__.__name__, mandatory_args
                )

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        return human_texts

    def _format_mask_token(
        self, idx: int, add_period: bool = False, add_whitespace: bool = True
    ) -> str:
        period = "." if add_period else ""
        whitespace = " " if add_whitespace else ""
        return f"{self.args['mask_token']}-{idx}{period}{whitespace}"

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        text_column = self.input_config.dataset_text_column
        texts = spacy_pipeline(
            texts=dataset[text_column], language=self.input_config.language
        )
        masked_texts = []
        for text in texts:
            words = [word.text_with_ws for word in text]
            percentage_masks = uniform(*self.args["percentage_range"])
            spans_to_mask = min(
                ceil((len(words) - 1) / max(self.args["span_length_range"])),
                ceil(
                    percentage_masks
                    * (len(words) / max(self.args["span_length_range"]))
                ),
            )

            sampled_positions: List[int] = []
            positions = list(range(len(words)))
            # Sample N words with at least max(args["span_length_range"])
            # positions of separation
            while (
                len(sampled_positions) < spans_to_mask
                and len(positions) >= spans_to_mask
            ):
                start_position = choice(positions)
                valid_positions = [
                    position
                    for position in positions
                    if abs(position - start_position)
                    >= max(self.args["span_length_range"])
                ]
                if valid_positions:
                    chosen_position = choice(valid_positions)
                    sampled_positions.append(chosen_position)
                    positions.remove(chosen_position)

            # Sample lengths and mask spans of the text
            sampled_positions = sorted(sampled_positions)
            span_lengths = [
                randint(*self.args["span_length_range"])
                for _ in range(len(sampled_positions))
            ]
            masked_words = []
            prev_position = 0
            mask_idxs = 0
            for position, length in zip(sampled_positions, span_lengths):
                masked_words.extend(words[prev_position:position])
                masked_words.append(self._format_mask_token(mask_idxs))
                prev_position = position + length
                mask_idxs += 1
            masked_texts.append("".join(masked_words))
        self.workspace["masked_texts"] = masked_texts
        return {"masked_text": masked_texts}



from enum import Enum
from typing import List

from pydantic import BaseModel, Field, model_validator


class DetectionLabels(Enum):
    """
    Labels for detection tasks.
    """

    GENERATED: str = "generated"
    HUMAN: str = "human"


class Placeholders(Enum):
    """
    Placeholders to be used across TextMachina.
    """

    NO_PROMPT: str = "NO-PROMPT"
    NO_EXTRACTOR: str = "NO-EXTRACTOR"


class Prompt(BaseModel):
    """
    Wrapper for a prompt.
    """

    template: str
    extractor: str


class PromptedDataset(BaseModel):
    """
    Wrapper for a prompted dataset used to generate MGT texts.
    """

    prompted_texts: List[str]
    human_texts: List[str]


class TaskType(str, Enum):
    DETECTION: str = "detection"
    ATTRIBUTION: str = "attribution"
    BOUNDARY: str = "boundary"
    MIXCASE: str = "mixcase"


class LabeledSpan(BaseModel):
    start: int = Field(ge=0)
    end: int
    label: str

    @model_validator(mode="after")
    def check_valid_positions(self) -> "LabeledSpan":
        if not (self.start < self.end):
            raise ValueError(
                "`start` must be lower than `end` in a LabeledSpan"
                f" (`start`={self.start}, `end`={self.end})"
            )
        return self



# flake8: noqa
from .config import Config, InputConfig, ModelConfig
from .constrainers import get_length_constrainer
from .data import PromptedDatasetBuilder
from .extractors import get_extractor
from .generators import get_generator
from .metrics import get_metric
from .models import get_model
from .tokenizers import get_tokenizer

__all__ = [
    "Config",
    "InputConfig",
    "ModelConfig",
    "get_length_constrainer",
    "PromptedDatasetBuilder",
    "get_extractor",
    "get_generator",
    "get_metric",
    "get_model",
    "get_tokenizer",
]



from typing import Dict, List, Set

import spacy
from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .types import EXTRACTOR_OMITTED
from .utils import spacy_pipeline


def extract_nouns(processed_text: spacy.tokens.Doc) -> Set[str]:
    """
    Extracts noun chunks from a Spacy doc.

    Args:
        processed_text (Doc): Spacy doc.

    Returns:
        Set[str]: noun chunks in the doc.
    """
    # take noun chunks -> nouns -> roots -> disregard
    nouns = [x.text for x in processed_text.noun_chunks]
    if not nouns:
        nouns = [x.text for x in processed_text if x.pos_ == "NOUN"]
    if not nouns:
        nouns = [x.text for x in processed_text if x.dep_ == "ROOT"]
    else:
        nouns = [EXTRACTOR_OMITTED]  # to filter after
    return set(nouns)


class NounList(Extractor):
    """
    Extractor that fills the prompt template with noun-phrases
    extracted from a text column in the dataset.

    This extractor needs a template placeholder named {nouns}.

    This extractor does not need specific arguments.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        processed_texts = spacy_pipeline(
            dataset[self.input_config.dataset_text_column],
            language=self.input_config.language,
            disable_pipes=[
                "ner",
                "senter",
                "attribute_ruler",
                "lemmatizer",
            ],
        )
        nouns = [", ".join(extract_nouns(text)) for text in processed_texts]
        return {"nouns": nouns}



import json
from pathlib import Path
from typing import Dict

import matplotlib.pyplot as plt
import mauve
from datasets import Dataset

from ..common.exceptions import InvalidTaskTypeForMetric, UnsupportedMetricParam
from ..types import TaskType
from .base import Metric


class MAUVEMetric(Metric):
    """
    Implements the MAUVE metric: https://arxiv.org/abs/2102.01454

    Supported tasks: detection.
    """

    def _run(self, dataset: Dataset, **kwargs) -> Dict:
        if self.task_type != TaskType.DETECTION:
            raise InvalidTaskTypeForMetric(self.name, self.task_type)

        self.check_kwargs_ok(kwargs)

        df = dataset.to_pandas()

        generated_texts = df[df["label"] == "generated"]["text"]
        human_texts = df[df["label"] == "human"]["text"]

        outputs = mauve.compute_mauve(
            p_text=generated_texts, q_text=human_texts, **kwargs
        )

        return vars(outputs)

    def _save(self, outputs: Dict, path: Path) -> None:
        plt.plot(
            outputs["divergence_curve"][:, 1], outputs["divergence_curve"][:, 0]
        )
        plt.savefig(path / "divergence_curve.pdf")

        result = {}
        for k, v in outputs.items():
            if k in {"mauve", "frontier_integral"}:
                result[k] = v
            elif k in {"q_hist", "p_hist"}:
                result[k] = v.tolist()

        with open(path / "summary.json", "w") as f:
            json.dump(result, f, indent=4)

    def _log(self, outputs: Dict, logger) -> None:
        logger.info(f"MAUVE score: {outputs['mauve']}")
        logger.info(f"Frontier Integral: {outputs['frontier_integral']}")

    def check_kwargs_ok(self, kwargs) -> None:
        # Can't accept these since they are mutually exclusive with passing text as input
        unsupported = [
            x
            for x in {"p_tokens", "q_tokens", "p_features", "q_features"}
            if x in kwargs
        ]
        if unsupported:
            raise UnsupportedMetricParam(unsupported[0], self.name)



from pathlib import Path

import evaluate
import pandas as pd
from datasets import Dataset

from ..common.exceptions import InvalidTaskTypeForMetric
from ..types import TaskType
from .base import Metric


class PerplexityMetric(Metric):
    """
    Implements the perplexity metric.

    Supported tasks: detection, attribution, and boundary.
    """

    def _run(self, dataset: Dataset, **kwargs) -> pd.DataFrame:
        metric = evaluate.load("perplexity", module_type="metric")

        # If we evaluate complete texts we can run easily grouping by labels
        if self.task_type in {TaskType.DETECTION, TaskType.ATTRIBUTION}:
            label = dataset["label"]
            text = dataset["text"]
        # Otherwise we need to separate the sections that pertain to different labels
        elif self.task_type == TaskType.BOUNDARY:
            df = dataset.to_pandas()[["text", "label"]]

            human_text = df.apply(
                lambda x: x["text"][: x["label"]], axis=1
            ).tolist()
            generated_text = df.apply(
                lambda x: x["text"][x["label"] :], axis=1
            ).tolist()

            text = human_text + generated_text
            label = ["human"] * len(human_text) + ["generated"] * len(
                generated_text
            )
        else:
            raise InvalidTaskTypeForMetric(self.name, self.task_type)

        result = metric.compute(predictions=text, **kwargs)

        df = pd.DataFrame(
            {"label": label, "perplexity": result["perplexities"]}
        )

        return df

    def _save(self, outputs: pd.DataFrame, path: Path) -> None:
        outputs.to_csv(path / "full_outputs.csv")
        outputs.drop("label", axis=1).mean(axis=0).to_json(
            path / "summary.json", indent=4
        )
        outputs.groupby("label").mean().reset_index().to_json(
            path / "per_label_summary.json", indent=4
        )

    def _log(self, outputs: pd.DataFrame, logger) -> None:
        summary = outputs.groupby("label").mean().reset_index()
        if self.task_type == TaskType.BOUNDARY:
            logger.info(f"Mean perplexity of segments: {summary.to_dict()}")
        else:
            logger.info(f"Mean perplexity: {summary.to_dict()}")



from math import ceil
from random import choice, sample, uniform
from typing import Any, Dict, List

from datasets import Dataset

from ..common import color_log, get_logger
from ..common.exceptions import ExtractorInvalidArgs
from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .utils import spacy_pipeline

_logger = get_logger(__name__)


class SentenceMasking(Extractor):
    """
    Extractor that fills the prompt template with a text with
    masked sentences and the LLM has to generate all the masked
    sentences.

    This extractor needs two template placeholders:
        - {masked_text}: will be filled with a text with masked sentences.

    This extractor allows to pass the following arguments in the
    `extractor_args` field from the config:
        - mask_token (str): mask token, e.g., "MASK". Several masks in a text
            will be appended with the index, e.g. "MASK-0"
        - percentage_range (List[float]): range delimiting the percentage
            of sentences to be masked. At least one sentence will be
            always masked.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        args: Dict[str, Any] = input_config.extractor_args.get(
            "sentence_masking", {}
        )
        workspace: Dict[str, Any] = {"masked_texts": []}
        super().__init__(input_config, task_type, workspace, args)

        _logger.warn(
            color_log(
                f"You are using the `{self.__class__.__name__}` extractor."
                " Consider that few models like GPT-4 can work properly with this"
                " type of generation. Models must be:\n"
                "1) Capable of generating proper JSON.\n"
                "2) Capable enough to generate all the masks.",
                "bold_yellow",
            )
        )

    def check_valid_args(self):
        mandatory_args = ["mask_token", "percentage_range"]
        for mandatory_arg in mandatory_args:
            if mandatory_arg not in self.args:
                raise ExtractorInvalidArgs(
                    self.__class__.__name__, mandatory_args
                )

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        return human_texts

    def _format_mask_token(
        self, idx: int, add_period: bool = False, add_whitespace: bool = True
    ) -> str:
        period = "." if add_period else ""
        whitespace = " " if add_whitespace else ""
        return f"{self.args['mask_token']}-{idx}{period}{whitespace}"

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        text_column = self.input_config.dataset_text_column
        texts = spacy_pipeline(
            texts=dataset[text_column], language=self.input_config.language
        )
        masked_texts = []
        for text in texts:
            sentences = [sent.text_with_ws for sent in text.sents]
            # In case of 1 sentences, avoid masking it,
            # just add a mask sentence to right or left randomly.
            if len(sentences) == 1:
                mask_position = choice([0, 1])
                fmt_masked_token = self._format_mask_token(0)
                sentences.insert(mask_position, fmt_masked_token)
                masked_texts.append("".join(sentences).strip())
            else:
                percentage_masks = uniform(*self.args["percentage_range"])
                # Min to avoid masking all the sentences.
                sents_to_mask = min(
                    len(sentences) - 1, ceil(percentage_masks * len(sentences))
                )
                positions = sorted(sample(range(len(sentences)), sents_to_mask))
                for mask_idx, sent_idx in enumerate(positions):
                    sentences[sent_idx] = self._format_mask_token(mask_idx)
                masked_texts.append("".join(sentences).strip())
        self.workspace["masked_texts"] = masked_texts
        return {"masked_text": masked_texts}



from typing import Dict, List, Set

import spacy
from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .types import EXTRACTOR_OMITTED
from .utils import spacy_pipeline


def extract_entities(processed_text: spacy.tokens.Doc) -> Set[str]:
    """
    Extracts entities from a Spacy doc.

    Args:
        processed_text (Doc): Spacy doc.

    Returns:
        Set[str]: named entities in the doc.
    """
    # take noun chunks -> nouns -> roots -> disregard
    entities = [x.text for x in processed_text.ents]
    if not entities:
        entities = [EXTRACTOR_OMITTED]  # to filter after
    return set(entities)


class EntityList(Extractor):
    """
    Extractor that fills the prompt template with entities
    extracted from a text column in the dataset.

    This extractor needs a template placeholder named {entities}.

    This extractor does not need specific arguments.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        processed_texts = spacy_pipeline(
            dataset[self.input_config.dataset_text_column],
            language=self.input_config.language,
            disable_pipes=[
                "senter",
                "attribute_ruler",
                "lemmatizer",
            ],
        )
        entities = [
            ", ".join(extract_entities(text)) for text in processed_texts
        ]
        return {"entities": entities}



import logging
import sys
from datetime import datetime
from pathlib import Path

_time = datetime.now()

COLORS = {
    "grey": "\x1b[38;20m",
    "yellow": "\x1b[33;20m",
    "bold_yellow": "\x1b[33;1m",
    "red": "\x1b[31;20m",
    "bold_red": "\x1b[31;1m",
    "reset": "\x1b[0m",
}


def color_log(text: str, color: str) -> str:
    """
    Add color to a log text.

    Args:
        text (str): a text.
        color (str): a color in `COLORS`

    Returns:
        str: a text with color codes added.
    """
    return COLORS[color] + text + COLORS["reset"]


def get_logger(module_name: str) -> logging.Logger:
    """
    Returns the logger used across TextMachina modules.

    Args:
        module_name (str): name of the module.

    Returns:
        logging.Logger: the logger.
    """
    logger = logging.getLogger(module_name)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(
        "[%(asctime)s] - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    logfile = (
        Path("logs")
        .joinpath(
            _time.strftime("%Y_%m_%d"),
            _time.strftime("%H_%M_%S"),
            "text-machina.log",
        )
        .absolute()
    )

    logfile.parent.mkdir(parents=True, exist_ok=True)

    fh = logging.FileHandler(logfile)
    fh.setFormatter(formatter)

    sh = logging.StreamHandler(sys.stdout)
    sh.setFormatter(formatter)

    logger.addHandler(sh)
    logger.addHandler(fh)

    return logger



from importlib import import_module
from pathlib import Path
from typing import Dict, List, Mapping

from datasets import Dataset

from ..common.exceptions import InvalidMetric, MissingMetricError
from ..types import TaskType
from .base import Metric

METRICS: Mapping[str, str] = {
    "mauve": "MAUVEMetric",
    "repetition_diversity": "RepetitionDiversityMetric",
    "simple_model": "SimpleModelMetric",
    "perplexity": "PerplexityMetric",
    "token_classification": "TokenClassificationMetric",
}


def run_metrics(
    dataset: Dataset,
    task_type: TaskType,
    save_dir: Path,
    metrics: List[str],
    metric_args: Dict[str, Dict],
) -> None:
    """
    Runs a list of metrics against a dataset.

    Args:
        dataset (Dataset): the dataset where metrics will be applied.
        task_type (TaskType): the type of task.
        save_dir (Path): the path where metric outputs will be saved.
        metrics: (List[str]): the list of metrics to run.
        metric_args: (Dict): the arguments of each metric
    """
    for name in metrics:
        metric = get_metric(task_type, name)
        metric.run(dataset, save_dir, **metric_args[name])


def get_metric(task_type: TaskType, name: str) -> Metric:
    """
    Gets a metric from the pool.

    Args:
        task_type (TaskType): the type of task.
        name (str): the metric name.
    Returns:
        Metric: the selected metric from the pool.

    """
    metric_cls_name = METRICS.get(name, None)
    if metric_cls_name is None:
        raise InvalidMetric(name)

    try:
        metric_class = getattr(
            import_module(f".{name}", __name__),
            metric_cls_name,
        )
    except (ModuleNotFoundError, ImportError):
        raise MissingMetricError(metric=name)

    return metric_class(task_type, name)


__all__ = list(METRICS.values())



from math import ceil
from random import randint
from typing import Any, Dict, List, Tuple

from datasets import Dataset

from ..common.exceptions import ExtractorInvalidArgs
from ..config import InputConfig
from ..types import TaskType
from .base import Extractor
from .utils import spacy_pipeline


class SentenceGap(Extractor):
    """
    Extractor that fills the prompt template with a boundary of
    two sentences (left-side and right-side of a sampled sentence),
    and with the number of sentences the LLM has to generate in
    between the boundary sentences.

    This extractor needs two template placeholders:
        - {n}: will be filled with the number of sentences to generate
               between the boundary sentences.
        - {boundaries}: will be filled with the boundary sentences separated
                        by the gap token and newlines.
                        E.g., "sentence 1.\n____\nsentence2"

    This extractor allows to pass the following arguments in the
    `extractor_args` field from the config:
        - gap_token (str): gap token, e.g., "____"
        - max_percentage_boundaries (float): max percentage of
          boundaries to sample from a text. In a text of N sentences,
          there will be N-1 possible boundaries of two sentences.
        - max_sentence_span (int): max number of sentences to be generated
          between the boundary sentences.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        args: Dict[str, Any] = input_config.extractor_args.get(
            "sentence_gap", {}
        )
        workspace: Dict[str, Any] = {
            "positions": [],
            "human_spans": [],
            "num_boundaries": [],
        }
        super().__init__(input_config, task_type, workspace, args)

    def check_valid_args(self):
        mandatory_args = [
            "gap_token",
            "max_percentage_boundaries",
            "max_sentence_span",
        ]
        for mandatory_arg in mandatory_args:
            if mandatory_arg not in self.args:
                raise ExtractorInvalidArgs(
                    self.__class__.__name__, mandatory_args
                )

    def prepare_human(self, human_texts: List[str]) -> List[str]:
        return [
            "".join(doc_sentences)
            for doc_sentences in self.workspace["human_spans"]
        ]

    def _format_boundary(self, pair: Tuple[str, str]) -> str:
        return f"{pair[0]}\n{self.args['gap_token']}\n{pair[1]}"

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        text_column = self.input_config.dataset_text_column
        texts = spacy_pipeline(
            texts=dataset[text_column], language=self.input_config.language
        )
        boundaries = []
        n_generated_sentences = []
        for text in texts:
            sentences = list([sent.text_with_ws for sent in text.sents])
            self.workspace["human_spans"].append(sentences)
            self.workspace["positions"].append([])
            # If the text has more than one sentence, it can be used
            # to interleave generations w/ human sentences.
            max_boundaries_to_select = ceil(
                (len(sentences) - 1) * self.args["max_percentage_boundaries"]
            )
            accum = 0
            if len(sentences) > 1:
                for idx in range(len(sentences) - 1):
                    is_boundary = randint(0, 1)
                    if is_boundary and accum < max_boundaries_to_select:
                        gen_sents = randint(1, self.args["max_sentence_span"])
                        boundary = (sentences[idx], sentences[idx + 1])
                        boundaries.append(self._format_boundary(boundary))
                        self.workspace["positions"][-1].append(idx)
                        n_generated_sentences.append(gen_sents)
                        accum += 1
            # At this point:
            # (1) No boundaries have been selected:
            #     -> consider all the text as human
            # (2) >=1 boundary have been selected:
            #     -> the text could be interleaved (gen, human)
            # (3) The text has only one sentence:
            #     -> consider all the text as human
            self.workspace["num_boundaries"].append(accum)

        return {
            "n": list(map(str, n_generated_sentences)),
            "boundaries": boundaries,
        }



# flake8: noqa
from typing import Mapping, Type

from ..config import Config
from ..types import TaskType
from .attribution import AttributionDatasetGenerator
from .base import DatasetGenerator
from .boundary import BoundaryDatasetGenerator
from .detection import DetectionDatasetGenerator
from .mixcase import MixCaseDatasetGenerator

GENERATORS: Mapping[TaskType, Type[DatasetGenerator]] = {
    TaskType.DETECTION: DetectionDatasetGenerator,
    TaskType.ATTRIBUTION: AttributionDatasetGenerator,
    TaskType.BOUNDARY: BoundaryDatasetGenerator,
    TaskType.MIXCASE: MixCaseDatasetGenerator,
}


def get_generator(config: Config) -> DatasetGenerator:
    """
    Gets a dataset generator from the pool.

    Args:
        config (Config): a config.

    Returns:
        DatasetGenerator: a dataset generator from the pool.
    """
    return GENERATORS[config.task_type](config)  # type: ignore


__all__ = [str(cls) for cls in GENERATORS.values()]



from typing import Dict, List

from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor


class Dummy(Extractor):
    """
    Dummy extractor that fills the prompt template with empty texts.

    This extractor needs one template placeholder named {dummy}.

    This extractor does not need specific arguments.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        return {
            "dummy": [
                ""
                for _ in range(
                    len(dataset[self.input_config.dataset_text_column])
                )
            ]
        }



import hashlib
import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pandas as pd
from datasets import (
    Dataset,
    DatasetDict,
    concatenate_datasets,
    load_dataset,
    load_from_disk,
)

from .common.logging import get_logger
from .config import Config, InputConfig
from .extractors import get_extractor
from .models.types import GENERATION_ERROR
from .tokenizers import get_tokenizer
from .types import Prompt, PromptedDataset, TaskType

_logger = get_logger(__name__)


class PromptedDatasetBuilder:
    """
    Class to manage all the prompting steps required before generating MGT.
    """

    def __init__(self, config: Config):
        self.config = config
        self.prompt = self.get_prompt()
        self.extractor = get_extractor(
            self.prompt.extractor, self.config.input, self.config.task_type
        )

    def build(self) -> PromptedDataset:
        """
        Prepares prefixes based on input formats for a particular
        domain, model and dataset.

        Returns:
            PromptedDataset: a dataset with prompted and human texts.
        """
        # load and prepare dataset
        dataset = load_dataset_from_config(self.config.input)

        # sample human texts
        human_texts, dataset = self.sampling(dataset)

        # compute prompt inputs and prepare human texts
        prompt_inputs = self.extractor.extract(dataset)
        human_texts = self.extractor.prepare_human(human_texts)

        # truncate the prompt inputs and format the prompts
        prompt_inputs = self.truncate_inputs(prompt_inputs)
        inputs = format_prompt(self.prompt.template, prompt_inputs)

        return PromptedDataset(prompted_texts=inputs, human_texts=human_texts)

    def sampling(self, dataset: Dataset) -> Tuple[List[str], Dataset]:
        """
        Sample human texts and texts to be used for generating MGT.
        The same amount is sampled in both cases.

        This method allows to randomly sample human texts, or use
        the same ones than those that will be used to generate MGT.

        Args:
            dataset (Dataset): a dataset.
        Returns:
            Tuple[List[str], Dataset]: tuple of texts. human texts and
                texts to be used to generate MGT.
        """
        dataset = dataset.shuffle()
        select_range = range(min(self.config.input.quantity, len(dataset)))

        # Disable random_sample_human automatically for boundary tasks
        if self.config.task_type == TaskType.BOUNDARY:
            _logger.info(
                "Automatically disabling `random_sample_human`"
                f"for the {TaskType.BOUNDARY.value} task."
            )
            self.config.input.random_sample_human = False

        if self.config.input.random_sample_human:
            human_texts = dataset.select(select_range)[
                self.config.input.dataset_text_column
            ]
            dataset = dataset.shuffle()
            dataset = dataset.select(select_range)
        else:
            dataset = dataset.select(select_range)
            human_texts = dataset[self.config.input.dataset_text_column]
        return human_texts, dataset

    def get_prompt(self) -> Prompt:
        """
        Returns the input format to be used as input for
        the text generation models

        Returns:
            Prompt: a prompt with template and extractor.
        """
        prompt = Prompt(
            template=self.config.input.template,
            extractor=self.config.input.extractor,
        )
        _logger.info(f"Prompt prepared: {prompt}")
        return prompt

    def truncate_inputs(
        self, prompt_inputs: Dict[str, List[str]]
    ) -> Dict[str, List[str]]:
        """
        Truncates prompt inputs extracted with the extractors.

        Args:
            prompt_inputs (Dict[str, List[str]]): prompt inputs.

        Returns:
            Dict[str, List[str]]: truncated prompt inputs.
        """

        max_input_tokens = self.config.input.max_input_tokens
        tokenizer = get_tokenizer(
            provider=self.config.model.provider,
            model_name=self.config.model.model_name,
        )

        prompt_inputs = tokenizer.distributed_truncate(
            prompt_inputs, max_input_tokens
        )

        _logger.info(f"Truncated prompt inputs to {max_input_tokens} tokens.")

        return prompt_inputs


def format_prompt(
    template: str, prompt_inputs: Dict[str, List[str]]
) -> List[str]:
    """
    Formats a prompt template with the prompt inputs.

    Example:
        template: "Write a text using this entities: {entities}.\n Text:"
        prompt_inputs: {"entities": ["Jose, Areg", "Marc, Angelo"]}
        output: ["Write a text using this entities: Jose, Areg.\n Text:",
                 "Write a text using this entities: Marc, Angelo.\n Text:"

    Args:
        template (str): the template to be formatted.
        prompt_inputs (Dict[str, List[str]]): prompt inputs from the extractors

    Returns:
        List[str]: formatted templates.
    """
    result = []
    input_keys = list(prompt_inputs.keys())
    for i in range(len(prompt_inputs[input_keys[0]])):
        args = {key: prompt_inputs[key][i] for key in input_keys}
        result.append(template.format(**args))
    return result


def serialize_dataset(
    dataset: Dataset,
    config: Config,
    path: Path,
    run_name: str,
) -> Path:
    """
    Saves a dataset with its config as an additional column.

    Args:
        dataset (Dataset): a dataset.
        config (Config): configuration used.
        path (Path): path where to save the generated dataset.
        run_name (str): name of this run.

    Returns:
        Path: folder where the dataset was saved.
    """
    save_path = get_save_path(config, path, run_name)

    save_path.mkdir(parents=True, exist_ok=True)

    dataset.save_to_disk(save_path.as_posix())
    _logger.info(f"The dataset has been saved in {save_path}")

    return save_path


def concatenate(paths: List[Path], save_path: Path) -> Dataset:
    """
    Concatenates and saves a list of datasets.

    Args:
        paths (List[Paths]): list with the datasets to be concatenated.
        save_path (Path): path where to save the merged dataset.

    Returns:
        Dataset: the merged dataset.
    """
    save_path.mkdir(parents=True, exist_ok=True)
    dataset = load_from_disk(paths[0].as_posix())

    for path in paths[1:]:
        dataset = concatenate_datasets(
            [dataset, load_from_disk(path.as_posix())]
        )

    return dataset


def load_dataset_from_config(config: InputConfig) -> Dataset:
    """
    Loads a dataset from disk or hub.

    Args:
        config (InputConfig): an input config.
    Returns:
        Dataset: a dataset.
    """
    try:
        dataset = load_from_disk(config.dataset)
        if "split" in config.dataset_params:
            dataset = dataset[config.dataset_params["split"]]
    except FileNotFoundError:
        dataset = load_dataset(config.dataset, **config.dataset_params)

    if isinstance(dataset, DatasetDict):
        split = list(dataset.keys())[0]
        _logger.warn(
            f"Picking the {split} split, since it was not"
            "specified in the config file."
        )
        dataset = dataset[split]

    return dataset


def get_save_path(
    config: Config, save_dir: Path, run_name: str, check_exists: bool = False
) -> Path:
    """
    Constructs the path to save a dataset.

    Args:
        config (Config): config of this run.
        save_dir (Path): root of the save path.
        run_name (str): name of this run.
        check_exists (bool): ...

    Returns:
        Path: path to save a dataset.

    """
    config_as_string = json.dumps(
        config.model_dump(), sort_keys=True, default=str
    ).encode()
    prefix = hashlib.sha256(config_as_string).hexdigest()

    parent = save_dir / run_name
    save_path = parent / prefix

    # check if path with prefix exists
    if check_exists:
        return get_path_from_substring(parent, prefix)  # type: ignore

    return save_path


def get_path_from_substring(path: Path, substring: str) -> Optional[Path]:
    """
    Checks whether a folder name within `path` includes `substring`

    Args:
        path (Path): path where searching for folders.
        substring (str): substring to find in the names.

    Returns:
        Optional[Path]: a path of a folder named *`substring`* or None.
    """
    for p in path.glob("*"):
        if substring in p.name:
            return p

    return None


def domain_model_counts(dataset: Dataset) -> pd.DataFrame:
    """
    Computes counts for (domain, model) pairs, e.g:

        model    bloom-560m  gpt2  human  total
        domain
        reviews          10    10     20     40
        tweets           10    10     20     40
        total            20    20     40     80

    Args:
        dataset (Dataset): the dataset used to compute counts.

    Returns:
        pd.DataFrame: the (domain, model) counts.
    """
    df = dataset.to_pandas()

    # We're interested in the number of available examples,
    # so we ignore the errors in the generation process.
    df = df[~df["text"].str.contains(GENERATION_ERROR)]

    counts = (
        df.groupby(["model", "domain"]).size().rename("count").reset_index()
    )

    counts = counts.pivot(columns="model", index="domain", values="count")

    counts.loc["total"] = counts.sum(axis=0)
    counts["total"] = counts.sum(axis=1)

    return counts


def errors_per_model(dataset: Dataset) -> pd.DataFrame:
    """
    Computes error counts per model.

    Args:
        dataset (Dataset): the dataset used to compute counts.

    Returns:
        pd.DataFrame: the error counts per model.
    """
    df = dataset.to_pandas()

    df["errors"] = df["text"].str.contains(GENERATION_ERROR)
    df["texts"] = ~df["errors"]

    counts = df[["model", "texts", "errors"]].groupby("model").sum()
    counts.loc["total"] = counts.sum(axis=0)

    return counts



import random
import re
from collections import defaultdict
from typing import Dict, List

from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor


class Example(Extractor):
    """
    Extractor that fills the prompt template with the text column
    of randomly sampled (w/o replacement) examples from the dataset.

    This extractor needs as many template placeholders named {example_i}
    as examples you want to include in the prompt.

    This extractor accepts the following arguments in the `extractor_args`
    field from the config:
        - all_random (bool): whether if the examples must be random
                             for each sample or not.
        - seed (int): the random seed.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)
        self.args = self.input_config.extractor_args.get("example", {})
        self.all_random = self.args.get("all_random", True)
        self.seed = self.args.get("seed", 13)

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        regex = r"\{(example_\d+)\}"
        example_placeholders = re.findall(regex, self.input_config.template)
        output = defaultdict(list)

        # Same examples for all the samples
        if not self.all_random:
            random_fixed = random.Random(self.seed)
            examples = dataset.select(
                random_fixed.sample(
                    range(len(dataset)), len(example_placeholders)
                )
            )[self.input_config.dataset_text_column]

            for idx, placeholder in enumerate(example_placeholders):
                output[placeholder] = [examples[idx]] * len(dataset)
        # Random examples for each sample
        else:
            for _ in range(len(dataset)):
                examples = dataset.select(
                    random.sample(
                        range(len(dataset)), len(example_placeholders)
                    )
                )[self.input_config.dataset_text_column]
                for idx, placeholder in enumerate(example_placeholders):
                    output[placeholder].append(examples[idx])

        return dict(output)



from .hf_local import HuggingFaceLocalTokenizer


class InferenceServerTokenizer(HuggingFaceLocalTokenizer):
    """
    Tokenizer for models deployed on inference servers
    like VLLM or TRT. This tokenizer assumes the tokenizers
    of the deployed models are available on the HF hub.
    """

    def __init__(self, model_name: str):
        super().__init__(model_name)



import os
from typing import List

from anthropic import Anthropic

from .base import Tokenizer


class AnthropicTokenizer(Tokenizer):
    """
    Tokenizer for Anthropic models.

    Requires the definition of the `ANTRHOPIC_API_KEY=<key>` environment variable.
    """

    def __init__(self, model_name: str):
        super().__init__(model_name)
        self.tokenizer = Anthropic(
            api_key=os.environ["ANTHROPIC_API_KEY"]
        ).get_tokenizer()

    def encode(self, text: str) -> List[int]:
        return self.tokenizer.encode(text).ids

    def decode(self, tokens: List[int]) -> str:
        return self.tokenizer.decode(tokens)



from typing import List

from datasets import Dataset

from ..common.exceptions import DatasetGenerationError
from ..config import Config
from .base import DatasetGenerator


class BoundaryDatasetGenerator(DatasetGenerator):
    """
    Dataset generator for the boundary task type.
    """

    def __init__(self, config: Config) -> None:
        super().__init__(config=config)

    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        """
        Combines and labels the generated and human texts.

        Args:
            generations (List[str]): list of generated texts.
            kwargs: additional keyword arguments.

        Returns:
            Dataset: a dataset including all the texts.
        """
        prompted_dataset = kwargs.get("prompted_dataset", None)
        if prompted_dataset is None:
            raise DatasetGenerationError(f"prompted_dataset not found: {self}")

        model_name = self.config.model.model_name
        domain = self.config.input.domain
        extractor = self.config.input.extractor

        dataset = Dataset.from_list(
            [
                {
                    "prompt": prompt,
                    "text": f"{human.strip()} {generated.strip()}",
                    "label": len(human.strip()),
                    "model": model_name,
                    "domain": domain,
                    "extractor": extractor,
                }
                for prompt, human, generated in zip(
                    prompted_dataset.prompted_texts,
                    prompted_dataset.human_texts,
                    generations,
                )
            ]
        )

        dataset = dataset.shuffle()

        return dataset



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from datasets import Dataset

from ..common.logging import get_logger
from ..types import TaskType

_logger = get_logger(__name__)


class Metric(ABC):
    """
    Base class for metrics.
    """

    def __init__(self, task_type: TaskType, name: str) -> None:
        self.name = name
        self.task_type = task_type

    def run(self, dataset: Dataset, path: Path, **kwargs) -> None:
        """Calls _run, _save and _log to compute a metric, store and log its outputs.

        Args:
            dataset (Dataset): A labeled dataset on which to compute the metric.
            path (Path): A path where to save results.
            **kwargs: Additional arguments to pass to _run.
        """
        outputs = self._run(dataset, **kwargs)

        path = path / self.name
        path.mkdir(parents=True, exist_ok=True)

        self._save(outputs, path)
        self._log(outputs, _logger)

    @abstractmethod
    def _run(self, dataset: Dataset, **kwargs) -> Any:
        """Implements metric logic.

        This method must implement all computation of the metric, and
        must be overriden in each new metric implementation, returning
        any result (ideally in raw form) that will be used in _save or _log.

        Args:
            dataset (Dataset): A labeled dataset on which to compute the metric.
            **kwargs: Additional arguments to use for metric computation.

        Returns:
            Any: The outputs of computing the metric on the dataset.

        Raises:
            InvalidTaskTypeForMetric: if a metric is ran on a
                task type its not defined for.
            UnsupportedMetricParam: if a parameter in kwargs is unsupported.
                See exception docs for more information.
        """
        ...

    @abstractmethod
    def _save(self, outputs: Any, path: Path) -> None:
        """Implements saving logic.

        Any operations related to saving metric outputs should
        be carried out here, including their processing for saving.
        Must be overriden in each new metric implementation.

        Any saving operations must be carried out in `path` or within
        its children, without having to access parent directories.

        Args:
            outputs (Any): The outputs of computing the metric with _run.
            path (Path): the path where to save the
        """
        ...

    @abstractmethod
    def _log(self, outputs: Any, logger: logging.Logger) -> None:
        """Implements logging logic.

        Any operations related to logging metric outputs should
        be carried out here, including their processing.
        Must be overriden in each new metric implementation.

        All logging operations must be done through the `logger` object.

        Args:
            outputs (Any): The outputs of computing the metric with _run.
            logger: (logging.Logger): The logger object to use for logging.
        """
        ...



import json
import os
from typing import Any, Dict

import boto3
from botocore import exceptions as boto_exceptions
from botocore.config import Config as BotoConfig

from ..common.logging import get_logger
from ..common.utils import get_instantiation_args
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR

_logger = get_logger(__name__)


class BedrockModel(TextGenerationModel):
    """
    Generates completions using AWS Bedrock models.

    Requires the definition of the `AWS_ACCESS_KEY_ID=<key>` and
    `AWS_SECRET_ACCESS_KEY=<key>` environment variables.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)
        client_config = BotoConfig(
            **get_instantiation_args(
                BotoConfig.__init__,
                self.model_config.model_dump(),
                accepted_params=list(BotoConfig.OPTION_DEFAULTS.keys()),
            )
        )
        self.client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
            config=client_config,
        )

    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict[str, Any],
    ) -> str:
        request_body = self.get_request_body(prompt, generation_config)

        try:
            response = self.client.invoke_model(
                body=request_body,
                modelId=self.model_config.model_name,
                accept="application/json",
                contentType="application/json",
            )
            response_body = json.loads(response["body"].read())
            completion = self.get_completion_from_response_body(response_body)

        except boto_exceptions.ClientError as ce:
            error_msg = ce.response["Error"]["Code"]
            _logger.info(
                f"Unrecoverable exception during the request: {error_msg}"
            )
            return GENERATION_ERROR
        return completion

    def get_request_body(
        self, prompt: str, generation_config: Dict[str, Any]
    ) -> str:
        """
        Prepares the request body for a request to a bedrock model.

        Considers the different parameters that each model provider accepts.

        Args:
            prompt (str): the prompt to use for generating text.
            generationc_config (Dict[str, Any]): the generation config.
        Returns:
            Dict: a serializable provider-specific request body.
        """
        bedrock_provider = self.model_config.model_name.split(".")[0]
        assert bedrock_provider in {
            "ai21",
            "amazon",
            "anthropic",
            "cohere",
            "meta",
            "mistral",
        }

        if bedrock_provider == "amazon":
            request_body = {
                "inputText": prompt,
                "textGenerationConfig": generation_config,
            }
        elif bedrock_provider == "ai21":
            if "maxTokenCount" in generation_config:
                generation_config["maxTokens"] = generation_config.pop(
                    "maxTokenCount"
                )
            request_body = {"prompt": prompt, **generation_config}
        elif bedrock_provider == "anthropic":
            if "maxTokenCount" in generation_config:
                generation_config["max_tokens_to_sample"] = (
                    generation_config.pop("maxTokenCount")
                )
            request_body = {"prompt": prompt, **generation_config}
        elif bedrock_provider == "cohere":
            # length constrainers work directly on providers themselves
            # so instead we overwrite the key name here
            if "maxTokenCount" in generation_config:
                generation_config["max_tokens"] = generation_config.pop(
                    "maxTokenCount"
                )
            request_body = {"prompt": prompt, **generation_config}
        elif bedrock_provider == "meta":
            if "maxTokenCount" in generation_config:
                generation_config["max_gen_len"] = generation_config.pop(
                    "maxTokenCount"
                )
            request_body = {"prompt": prompt, **generation_config}
        elif bedrock_provider == "mistral":
            if "maxTokenCount" in generation_config:
                generation_config["max_tokens"] = generation_config.pop(
                    "maxTokenCount"
                )
            request_body = {"prompt": prompt, **generation_config}

        return json.dumps(request_body)

    def get_completion_from_response_body(self, response_body: Dict) -> str:
        """
        Obtains the completions from a response body returned by a bedrock model.

        Considers the different API schemas that each model provider uses.

        Args:
            response_body (Dict): the body returned by models in bedrock.
        Returns:
            str: the completion of the model extracted from the body.
        """
        bedrock_provider = self.model_config.model_name.split(".")[0]
        assert bedrock_provider in {
            "ai21",
            "amazon",
            "anthropic",
            "cohere",
            "meta",
            "mistral",
        }

        if bedrock_provider == "amazon":
            completion = response_body["results"][0]["outputText"]
        elif bedrock_provider == "ai21":
            completion = response_body["completions"][0]["data"]["text"]
        elif bedrock_provider == "anthropic":
            completion = response_body["completion"]
        elif bedrock_provider == "cohere":
            completion = response_body["generations"][0]["text"]
        elif bedrock_provider == "meta":
            completion = response_body["generation"]
        elif bedrock_provider == "mistral":
            completion = response_body["outputs"][0]["text"]

        return completion



# flake8: noqa
from importlib import import_module
from typing import Mapping

from ..common.exceptions import MissingIntegrationError
from ..config import ModelConfig
from .base import TextGenerationModel

MODELS: Mapping[str, str] = {
    "openai": "OpenAIModel",
    "anthropic": "AnthropicModel",
    "cohere": "CohereModel",
    "hf_local": "HuggingFaceLocalModel",
    "hf_remote": "HuggingFaceRemoteModel",
    "vertex": "VertexModel",
    "bedrock": "BedrockModel",
    "ai21": "AI21Model",
    "azure_openai": "AzureOpenAIModel",
    "inference_server": "InferenceServerModel",
}


def get_model(model_config: ModelConfig) -> TextGenerationModel:
    """
    Gets a text generation model from the pool.

    Args:
        model_config (ModelConfig): a model config.

    Returns:
        TextGenerationModel: a text generation model from the pool.
    """
    provider = model_config.provider
    try:
        model_class = getattr(
            import_module(f".{provider}", __name__),
            MODELS[provider],
        )
    except (ModuleNotFoundError, ImportError):
        raise MissingIntegrationError(integration=provider)
    return model_class(model_config)


__all__ = list(MODELS.values())



from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
import sklearn.metrics
import textstat
from datasets import Dataset
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.pipeline import FeatureUnion, Pipeline, make_pipeline
from tqdm import tqdm

from ..common.exceptions import InvalidTaskTypeForMetric
from ..types import TaskType
from .base import Metric


class SimpleModelMetric(Metric):
    """
    Implements simple baseline models evaluated with
    stratified k-fold validation.

    Supported tasks: detection, attribution, and boundary.
    """

    def _run(self, dataset: Dataset, **kwargs) -> Dict[int, pd.DataFrame]:
        if self.task_type not in {
            TaskType.DETECTION,
            TaskType.ATTRIBUTION,
            TaskType.BOUNDARY,
        }:
            raise InvalidTaskTypeForMetric(self.name, self.task_type)

        df = dataset.to_pandas()

        if self.task_type == TaskType.BOUNDARY:
            kf = KFold(**kwargs.get("folds", {}))
        else:
            kf = StratifiedKFold(**kwargs.get("folds", {}))

        X, y = df["text"], df["label"]

        reports = {}
        iterator = tqdm(
            enumerate(kf.split(df, df["label"])),
            desc="Running simple model",
            total=kf.get_n_splits(),
        )

        for i, (train_idx, test_idx) in iterator:
            train_X, train_y = X.iloc[train_idx], y.iloc[train_idx]
            test_X, test_y = X.iloc[test_idx], y.iloc[test_idx]

            # Predict based on max readability differences for every prefix-suffix pair
            if self.task_type == TaskType.BOUNDARY:
                preds = self._predict_for_boundaries(test_X)
                report = regression_report(test_y, preds)
            # Train a BOW+BOC+LR model
            else:
                model = self._get_model(kwargs)
                model.fit(train_X, train_y)
                preds = model.predict(test_X)
                report = pd.DataFrame(
                    sklearn.metrics.classification_report(
                        test_y, preds, output_dict=True, zero_division=0.0
                    )
                )
            reports[i] = report

        return reports

    def _save(self, outputs: Dict[int, pd.DataFrame], path: Path) -> None:
        single_path = path / "single"
        single_path.mkdir(parents=True, exist_ok=True)

        for k, df in outputs.items():
            df.to_json(single_path / f"{k}_single.json", indent=4)
            df.to_markdown(single_path / f"{k}_single.md")

        aggregated = self._get_aggregated_results(outputs, std=True)

        aggregated.to_json(path / "aggregated.json", indent=4)
        aggregated.to_markdown(path / "aggregated.md")

    def _log(self, outputs: Dict[int, pd.DataFrame], logger) -> None:
        aggregated = self._get_aggregated_results(outputs, std=False)
        logger.info(f"Mean results:\n {aggregated}")

    def _get_aggregated_results(
        self, outputs: Dict[int, pd.DataFrame], std: bool = False
    ) -> pd.DataFrame:
        concatenated = pd.concat(outputs.values())

        if std:
            aggregated = (
                concatenated.reset_index().groupby("index").agg(["mean", "std"])
            )
            aggregated.columns = [
                "_".join(col) for col in aggregated.columns.values
            ]
        else:
            aggregated = concatenated.reset_index().groupby("index").mean()

        # Support must be added, not averaged
        totals = (
            pd.concat(outputs.values()).reset_index().groupby("index").sum()
        )
        if self.task_type == TaskType.BOUNDARY:
            if std:
                aggregated.loc[0, "support_mean"] = totals["support"].iloc[0]
                aggregated.loc[0, "support_std"] = 0.0
            else:
                aggregated.loc[0, "support"] = totals["support"].iloc[0]
        else:
            labels = [
                x
                for x in outputs[0].columns
                if x not in {"macro avg", "weighted avg", "accuracy"}
            ]
            for label in labels:
                if std:
                    aggregated.loc["support", f"{label}_mean"] = totals.loc[
                        "support"
                    ][label]
                    aggregated.loc["support", f"{label}_std"] = 0.0
                else:
                    aggregated.loc["support", label] = totals.loc["support"][
                        label
                    ]

        return aggregated

    def _get_model(self, kwargs) -> Pipeline:
        char_params = kwargs.get("feature_params", {}).get("char", {})
        word_params = kwargs.get("feature_params", {}).get("word", {})
        model_params = kwargs.get("model_params", {})

        return make_pipeline(
            FeatureUnion(
                [
                    ("char", CountVectorizer(analyzer="char", **char_params)),
                    (
                        "word",
                        CountVectorizer(analyzer="word", **word_params),
                    ),
                ]
            ),
            LogisticRegression(**model_params),
        )

    def _predict_for_boundaries(self, texts: List[str]) -> np.ndarray:
        """Predict position of maximal difference in readabilities"""
        preds = []
        for text in texts:
            result = []
            N = len(text)
            # prefixes and suffixes every 10 characters
            # running this without jumps would be too slow
            for i in range(20, N - 20, 10):
                prefix, suffix = text[:i], text[i:]
                prefix_score = textstat.flesch_reading_ease(prefix)
                suffix_score = textstat.flesch_reading_ease(suffix)

                # distance linearly weighed by index
                # weighing is necessary since otherwise the maximal difference
                # will always be between first prefix and suffix due to their
                # differences in length
                if i > N // 2:
                    weight = 2 - (i / (N // 2))
                else:
                    weight = i / (N // 2)

                result.append(abs(prefix_score - suffix_score) * weight)

            preds.append(result.index(max(result)))

        return np.array(preds)


def regression_report(y_true: np.ndarray, y_pred: np.ndarray) -> pd.DataFrame:
    """
    Computes a regression report similarly to the classification report
    offered by scikit-learn.

    Args:
        y_true (np.ndarray): the true labels.
        y_pred (np.ndarray): the predicted labels.
    Returns:
        pd.DataFrame: the report.
    """
    metrics = {
        "mean absolute error": [
            sklearn.metrics.mean_absolute_error(y_true, y_pred)
        ],
        "median absolute error": [
            sklearn.metrics.median_absolute_error(y_true, y_pred)
        ],
        "mean squared error": [
            sklearn.metrics.mean_squared_error(y_true, y_pred)
        ],
        "max error": [sklearn.metrics.max_error(y_true, y_pred)],
        "r2 score": [sklearn.metrics.r2_score(y_true, y_pred)],
        "explained variance score": [
            sklearn.metrics.explained_variance_score(y_true, y_pred)
        ],
        "support": [len(y_true)],
    }

    df = pd.DataFrame(metrics)
    return df



import os
from typing import Dict

from cohere import Client

from ..common.logging import get_logger
from ..common.utils import get_instantiation_args
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR

_logger = get_logger(__name__)


class CohereModel(TextGenerationModel):
    """
    Generates completions using Cohere models.

    Requires the definition of the `COHERE_API_KEY=<key>` environment variable.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)
        self.client = Client(
            api_key=os.environ["COHERE_API_KEY"],
            **get_instantiation_args(
                Client.__init__, self.model_config.model_dump()
            ),
        )

    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict,
    ) -> str:
        try:
            completion = (
                self.client.generate(
                    model=self.model_config.model_name,
                    prompt=prompt,
                    **generation_config,
                )
                .generations[0]
                .text
            )
        except Exception as e:
            _logger.info(f"Unrecoverable exception during the request: {e}")
            return GENERATION_ERROR
        return completion



import os
from typing import Dict

from google.cloud import aiplatform
from google.oauth2 import service_account
from vertexai.preview.language_models import ChatModel as VertexChatModel
from vertexai.preview.language_models import (
    TextGenerationModel as VertexTextGenerationModel,
)

from ..common.logging import get_logger
from ..common.utils import get_instantiation_args
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR, CompletionType

_logger = get_logger(__name__)


class VertexModel(TextGenerationModel):
    """
    Generates completions using VertexAI models.
    Requires the definition of the `VERTEX_AI_CREDENTIALS_FILE=<path>` environment variable.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)

        credentials = service_account.Credentials.from_service_account_file(
            os.environ["VERTEX_AI_CREDENTIALS_FILE"]
        )
        aiplatform.init(
            credentials=credentials,
            **get_instantiation_args(
                aiplatform.init, self.model_config.model_dump()
            ),
        )
        self.model = self._get_model()

    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict,
    ) -> str:
        completion_fn = (
            self._chat_request
            if self.model_config.api_type == CompletionType.CHAT
            else self._completion_request
        )
        try:
            completion = completion_fn(prompt, generation_config)
        except Exception as e:
            _logger.info(f"Unrecoverable exception during the request: {e}")
            return GENERATION_ERROR
        return completion

    def _get_model(self):
        model_class = (
            VertexChatModel
            if self.model_config.api_type == CompletionType.CHAT
            else VertexTextGenerationModel
        )
        return model_class.from_pretrained(self.model_config.model_name)

    def _chat_request(self, prompt: str, generation_config: Dict) -> str:
        return (
            self.model.start_chat()
            .send_message(prompt, **generation_config)
            .text
        )

    def _completion_request(self, prompt: str, generation_config: Dict) -> str:
        return self.model.predict(prompt, **generation_config).text



from typing import Any, Dict, List

from datasets import Dataset
from rich.markup import escape
from rich.panel import Panel

from .base import Explorer


class MixCaseExplorer(Explorer):
    """
    Explorer for mixcase tasks.
    """

    def __init__(self, dataset: Dataset) -> None:
        super().__init__(dataset)

    def get_panels(self, example: Dict[str, Any]) -> List[Panel]:
        panels = []
        # Prompt panels
        prompts = (
            [example["prompt"]]
            if not isinstance(example["prompt"], list)
            else example["prompt"]
        )
        for idx, prompt in enumerate(prompts):
            panels.append(
                Panel(
                    escape(prompt),
                    title=f"[red3]Prompt {idx}",
                    border_style="turquoise2",
                    style="white",
                )
            )
        # Text panel
        text = ""
        for label in example["label"]:
            start, end = label["start"], label["end"]
            color = "[red3]" if label["label"] == "human" else "[blue3]"
            text += color + escape(example["text"][start:end])

        model = example["model"]
        title = (
            "[red3]Human[/] text"
            if model == "human"
            else f"Mixcase text from [blue3]{model}[/] and [red3]human[/]"
        )
        panels.append(
            Panel(
                text,
                title=title,
                border_style="turquoise2",
                style="white",
            )
        )
        return panels

    def get_title(self, idx: int, example: Dict[str, Any]) -> str:
        example["model"]
        domain = example["domain"]
        extractor = example["extractor"]

        title = f"[tan]{idx}[/]: "
        title += f"Mixcase text in domain [sea_green3]{domain}[/]"
        title += f" using extractor [yellow]{extractor}"

        return title



from typing import Any, Dict

import requests
from requests.adapters import HTTPAdapter, Retry

from ..common.exceptions import InvalidInferenceServer
from ..common.logging import get_logger
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR

_logger = get_logger(__name__)


ALLOWED_INFERENCE_SERVERS = ["vllm", "trt"]


class InferenceServerModel(TextGenerationModel):
    """
    Generates completions using models deployed
    on inference servers like TRT or VLLM. This
    model assumes the default APIs are being used
    (e.g., no OpenAI-compatible API in VLLM)
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)

        if self.model_config.inference_server not in ALLOWED_INFERENCE_SERVERS:
            raise InvalidInferenceServer(
                self.model_config.inference_server, ALLOWED_INFERENCE_SERVERS
            )

        self.base_url = self.model_config.base_url
        self.client = requests.Session()
        retry_adapter = HTTPAdapter(
            max_retries=Retry(
                total=getattr(self.model_config, "max_retries", 5),
                backoff_factor=getattr(
                    self.model_config, "backoff_factor", 0.5
                ),
                status_forcelist=[
                    code for code in requests.status_codes._codes if code != 200
                ],
            )
        )
        self.client.mount("http://", retry_adapter)
        self.client.mount("https://", retry_adapter)

    def prepare_data(
        self, prompt: str, generation_config: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Prepares the json data of a request to the inference server.

        Args:
            prompt (str): a prompt.
            generation_config (Dict[str, Any]): a generation config.

        Returns:
            Dict[str, Any]: json data for a request.
        """
        if self.model_config.inference_server == "vllm":
            return {"text_input": prompt, "parameters": generation_config}
        else:
            return {**{"text_input": prompt}, **generation_config}

    def parse_response(self, response: Dict, prompt_len: int) -> str:
        """
        Get the completion from a response, removing the
        prompt from it if required.

        Args:
            response (Dict): response from the inference server.
            prompt_len (int): len of the prompt.

        Returns:
            str: completion.
        """
        if self.model_config.inference_server == "vllm":
            return response["text_output"][prompt_len:]
        else:
            return response["text_output"][prompt_len:]

    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict,
    ) -> str:
        json_data = self.prepare_data(prompt, generation_config)
        try:
            response = self.client.post(self.base_url, json=json_data).json()
            completion = self.parse_response(response, len(prompt))
        except Exception as e:
            _logger.info(f"Unrecoverable exception during the request: {e}")
            return GENERATION_ERROR
        return completion



import json
import re
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Type

from datasets import Dataset, concatenate_datasets

from ..common import color_log, get_logger
from ..common.exceptions import DatasetGenerationError
from ..config import Config
from ..extractors import (
    Extractor,
    SentenceGap,
    SentenceMasking,
    WordGap,
    WordMasking,
)
from ..models.types import GENERATION_ERROR
from ..types import DetectionLabels, LabeledSpan, Placeholders
from .base import DatasetGenerator

_logger = get_logger(__name__)


class MixCaseDatasetGenerator(DatasetGenerator):
    """
    Dataset generator for the mixcase task type.
    """

    def __init__(self, config: Config) -> None:
        super().__init__(config=config)

    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        if isinstance(self.prompter.extractor, (SentenceGap, WordGap)):
            packer: Type[MixCasePacker] = MixCaseGapPacker
        elif isinstance(
            self.prompter.extractor, (SentenceMasking, WordMasking)
        ):
            packer = MixCaseMaskPacker
        else:
            packer = MixCaseRewritingPacker

        return packer(self.config, self.prompter.extractor)._pack(
            generations, **kwargs
        )


class MixCasePacker(ABC):
    """
    Base class for mixcase packers.
    """

    def __init__(self, config: Config, extractor: Extractor) -> None:
        self.config = config
        self.extractor = extractor

    @abstractmethod
    def _build_samples(
        self,
        generations: List[str],
    ) -> Tuple[List[str], List[List[Dict]]]:
        """
        Builds the samples to be packed in a dataset by preparing the texts
        and the labels according to the extractor (gap-based or mask-based).

        Args:
            generations (List[str]): list of generations.
            kwargs (Dict): additional keyword arguments.

        Returns:
            Tuple[List[str], List[List[Dict]]]: texts and labels to be added
                to the dataset.
        """
        ...

    @abstractmethod
    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        """
        Combines and labels the generated and human texts
        according to the extractor.

        Args:
            generations (List[str]): list of generated texts.
            kwargs: additional keyword arguments.

        Returns:
            Dataset: a dataset including all the texts.
        """
        ...


class MixCaseRewritingPacker(MixCasePacker):
    """
    Packer for mixcase task type when using rewriting-based extractors.
    """

    def __init__(self, config: Config, extractor: Extractor) -> None:
        super().__init__(config=config, extractor=extractor)

    def _build_samples(
        self,
        generations: List[str],
    ) -> Tuple[List[str], List[List[Dict]]]:
        """
        Interleaves generated and human spans to build mixcase samples.
        The `start` and `end` of the labels follow the Python [`start`, `end`)
        convention, i.e., including the `start` element and excluding `end`.
        For instance, given the text: "I like Apolo. I don't like Athenea",
        being the first sentence human-written and the second one generated,
        the labels will be:
        [
            {"start": 0, "end": 14, "label": "human"},
            {"start": 14, "end": 34, "label": "generated"},
        ]

        Args:
            generations (List[str]): list of generations

        Returns:
            Tuple[List[str], List[List[Dict]]]: the interleaved texts and
            the list of labels of each text.
        """
        texts, labels = [], []

        prev_sample = 0
        for sentences, positions in zip(
            self.extractor.workspace["human_spans"],
            self.extractor.workspace["positions"],
        ):
            rewrittens = generations[prev_sample : prev_sample + len(positions)]
            positions = sorted(positions)
            sample_labels: List[Dict] = []

            # Replace rewritten sentences
            sentences_with_rewrites = sentences[:]
            for idx, position in enumerate(positions):
                sentences_with_rewrites[position] = rewrittens[idx]
            sample_text = "".join(sentences_with_rewrites)

            # Mark generated spans
            generated_labels = []
            for idx, position in enumerate(positions):
                start_pos = len("".join(sentences_with_rewrites[:position]))
                end_pos = len("".join(sentences_with_rewrites[: position + 1]))
                generated_labels.append(
                    LabeledSpan(
                        start=start_pos,
                        end=end_pos,
                        label=DetectionLabels.GENERATED.value,
                    ).model_dump()
                )

            # Mark human spans
            human_labels = []
            prev_generated_position = 0
            for idx in range(len(generated_labels)):
                generated_label = generated_labels[idx]
                if prev_generated_position < generated_label["start"]:
                    human_labels.append(
                        LabeledSpan(
                            start=prev_generated_position,
                            end=generated_label["start"],
                            label=DetectionLabels.HUMAN.value,
                        ).model_dump()
                    )
                prev_generated_position = generated_label["end"]

            # Add human label if the end of the text has not been
            # reached by the last generated span
            last_generated_position = generated_labels[-1]["end"]
            if last_generated_position < len(sample_text):
                human_labels.append(
                    LabeledSpan(
                        start=last_generated_position,
                        end=len(sample_text),
                        label=DetectionLabels.HUMAN.value,
                    ).model_dump()
                )

            # Join generated labels with human labels and merge overlappings.
            merged_labels = sorted(
                human_labels + generated_labels, key=lambda span: span["start"]
            )
            sample_labels = []
            while len(merged_labels):
                current_span = merged_labels.pop(0)
                current_label = current_span["label"]
                spans_with_same_label = []
                while (
                    len(merged_labels)
                    and merged_labels[0]["label"] == current_label
                ):
                    spans_with_same_label.append(merged_labels.pop(0))

                if not spans_with_same_label:
                    sample_labels.append(current_span)
                else:
                    sample_labels.append(
                        LabeledSpan(
                            start=current_span["start"],
                            end=spans_with_same_label[-1]["end"],
                            label=current_label,
                        ).model_dump()
                    )

            prev_sample += len(positions)
            texts.append("".join(sentences_with_rewrites))
            labels.append(sample_labels)
        return texts, labels

    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        """
        Combines and labels the generated and human texts
        when using rewriting-based extractors (`sentence_rewriting`)

        Args:
            generations (List[str]): list of generated texts.
            kwargs: additional keyword arguments.

        Returns:
            Dataset: a dataset including all the texts.
        """
        prompted_dataset = kwargs.get("prompted_dataset", None)
        if prompted_dataset is None:
            raise DatasetGenerationError(f"prompted_dataset not found: {self}")

        model_name = self.config.model.model_name
        domain = self.config.input.domain
        extractor_name = self.config.input.extractor
        texts, labels = self._build_samples(generations)

        prev_sample = 0
        mixed_samples = []
        for idx, (text, sample_labels) in enumerate(zip(texts, labels)):
            n_positions = len(self.extractor.workspace["positions"][idx])
            prompt = prompted_dataset.prompted_texts[
                prev_sample : prev_sample + n_positions
            ] or [Placeholders.NO_PROMPT.value]

            mixed_samples.append(
                {
                    "prompt": prompt,
                    "text": text,
                    "label": sample_labels,
                    "model": model_name,
                    "domain": domain,
                    "extractor": extractor_name,
                }
            )
            prev_sample += n_positions

        mixed_dataset = Dataset.from_list(mixed_samples)

        human_dataset = Dataset.from_list(
            [
                {
                    "prompt": [Placeholders.NO_PROMPT.value],
                    "text": text,
                    "label": [
                        LabeledSpan(
                            start=0,
                            end=len(text),
                            label=DetectionLabels.HUMAN.value,
                        ).model_dump()
                    ],
                    "model": DetectionLabels.HUMAN.value,
                    "domain": domain,
                    "extractor": Placeholders.NO_EXTRACTOR.value,
                }
                for text in prompted_dataset.human_texts
            ]
        )

        dataset = concatenate_datasets([human_dataset, mixed_dataset])
        dataset = dataset.shuffle()

        return dataset


class MixCaseMaskPacker(MixCasePacker):
    """
    Packer for mixcase task type when using mask-based extractors.
    """

    def __init__(self, config: Config, extractor: Extractor) -> None:
        super().__init__(config=config, extractor=extractor)
        self.mask_regex = re.compile(
            rf"({self.extractor.args['mask_token']}-\d+)"
        )

    def _build_error_sample(self) -> Tuple[str, List[Dict]]:
        """
        Helper to build an error sample when required, e.g.,
        invalid JSON, missing masks, etc; that logs the cause of
        the resulting error (LLM uncapable always).

        Returns:
            Tuple[str, List[Dict]]: the text and labels of an error sample.
        """
        _logger.info(
            color_log(
                "The completion was not a valid JSON."
                " This error is related to the LLM capabilities,"
                " please, use another LLM if there are many"
                " errors of this type."
                " The text will be considered a generation error"
                f"`{GENERATION_ERROR}`",
                "bold_yellow",
            )
        )
        return GENERATION_ERROR, [
            LabeledSpan(
                start=0,
                end=len(GENERATION_ERROR),
                label=DetectionLabels.GENERATED.value,
            ).model_dump()
        ]

    def _build_sample(
        self, masked_text: str, parsed_generation: Dict[str, str]
    ) -> Tuple[str, List[Dict]]:
        """
        Builds a sample by reconstructing the masks in a text.

        Args:
            masked_text (str): a text with masks to be replaced.
            parsed_generation: (Dict[str, str]): dictionary mapping masks
                to texts, e.g. {"MASK-0": <text>, ...}

        Returns:
            Tuple[str, List[Dict]]: the text and labels of the sample.
        """
        # Replace masks in text and compute the generated labels
        # The following algorithm relies on masks sorted by their index.
        mask_completions = sorted(
            parsed_generation.items(), key=lambda x: int(x[0].split("-")[1])
        )
        sample_text = masked_text
        generated_labels = []
        for mask_token, completion in mask_completions:
            mask_match = re.search(rf"{mask_token}\b", sample_text)
            if mask_match is None:
                return self._build_error_sample()
            mask_position = mask_match.span()[0]
            # Ensure the completion does not append another mask token
            completion = self.mask_regex.sub("", completion)
            sample_text = re.sub(rf"({mask_token})\b", completion, sample_text)
            gen_end = mask_position + len(completion)
            generated_labels.append(
                LabeledSpan(
                    start=mask_position,
                    end=gen_end + 1,
                    label=DetectionLabels.GENERATED.value,
                ).model_dump()
            )

        # Compute human labels
        human_labels = []
        prev_generated_position = 0
        for idx in range(len(generated_labels)):
            generated_label = generated_labels[idx]
            if prev_generated_position < generated_label["start"]:
                human_labels.append(
                    LabeledSpan(
                        start=prev_generated_position,
                        end=generated_label["start"],
                        label=DetectionLabels.HUMAN.value,
                    ).model_dump()
                )
            prev_generated_position = generated_label["end"]

        # Add human label if the end of the text has not been
        # reached by the last generated span
        last_generated_position = generated_labels[-1]["end"]
        if last_generated_position < len(sample_text):
            human_labels.append(
                LabeledSpan(
                    start=last_generated_position,
                    end=len(sample_text),
                    label=DetectionLabels.HUMAN.value,
                ).model_dump()
            )

        # Join generated labels with human labels and merge overlappings.
        merged_labels = sorted(
            human_labels + generated_labels, key=lambda span: span["start"]
        )
        sample_labels = []
        while len(merged_labels):
            current_span = merged_labels.pop(0)
            current_label = current_span["label"]
            spans_with_same_label = []
            while (
                len(merged_labels)
                and merged_labels[0]["label"] == current_label
            ):
                spans_with_same_label.append(merged_labels.pop(0))

            if not spans_with_same_label:
                sample_labels.append(current_span)
            else:
                sample_labels.append(
                    LabeledSpan(
                        start=current_span["start"],
                        end=spans_with_same_label[-1]["end"],
                        label=current_label,
                    ).model_dump()
                )
        return sample_text, sample_labels

    def _build_samples(
        self,
        generations: List[str],
    ) -> Tuple[List[str], List[List[Dict]]]:
        """
        Reconstructs masked texts using completions to build mixcase samples.
        The `start` and `end` of the labels follow the Python [`start`, `end`)
        convention, i.e., including the `start` element and excluding `end`.
        For instance, given the text: "I like Apolo. I don't like Athenea",
        being the first sentence human-written and the second one generated,
        the labels will be:
        [
            {"start": 0, "end": 14, "label": "human"},
            {"start": 14, "end": 34, "label": "generated"},
        ]

        Args:
            generations (List[str]): list of generations

        Returns:
            Tuple[List[str], List[List[Dict]]]: the interleaved texts and
            the list of labels of each text.
        """
        texts, labels = [], []
        for generation, masked_text in zip(
            generations, self.extractor.workspace["masked_texts"]
        ):
            try:
                parsed_generation = json.loads(
                    generation[
                        generation.find("{") : generation.rfind("}") + 1
                    ].strip()
                )
            except json.decoder.JSONDecodeError:
                # If the LLM didn't generate a valid JSON output,
                # the sample will be considered as an error.
                sample_text, sample_labels = self._build_error_sample()
            else:
                # If the number of masks in the json do not match with
                # the number of masks in the text, the sample will be
                # considered as an error.
                masks_in_text = set(self.mask_regex.findall(masked_text))
                masks_in_completion = set(parsed_generation.keys())

                if len(masks_in_text.intersection(masks_in_completion)) != len(
                    masks_in_text
                ):
                    sample_text, sample_labels = self._build_error_sample()
                # Otherwise, the sample can be built.
                else:
                    sample_text, sample_labels = self._build_sample(
                        masked_text, parsed_generation
                    )
            finally:
                texts.append(sample_text)
                labels.append(sample_labels)
        return texts, labels

    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        """
        Combines and labels the generated and human texts
        when using mask-based extractors (`sentence_masking`, `word_masking`, etc.)

        Args:
            generations (List[str]): list of generated texts.
            kwargs: additional keyword arguments.

        Returns:
            Dataset: a dataset including all the texts.
        """
        prompted_dataset = kwargs.get("prompted_dataset", None)
        if prompted_dataset is None:
            raise DatasetGenerationError(f"prompted_dataset not found: {self}")

        model_name = self.config.model.model_name
        domain = self.config.input.domain
        extractor_name = self.config.input.extractor
        texts, labels = self._build_samples(generations)

        generated_dataset = Dataset.from_list(
            [
                {
                    "prompt": prompt,
                    "text": text,
                    "label": label,
                    "model": model_name,
                    "domain": domain,
                    "extractor": extractor_name,
                }
                for prompt, text, label in zip(
                    prompted_dataset.prompted_texts, texts, labels
                )
            ]
        )

        human_dataset = Dataset.from_list(
            [
                {
                    "prompt": Placeholders.NO_PROMPT.value,
                    "text": text,
                    "label": [
                        LabeledSpan(
                            start=0,
                            end=len(text),
                            label=DetectionLabels.HUMAN.value,
                        ).model_dump()
                    ],
                    "model": DetectionLabels.HUMAN.value,
                    "domain": domain,
                    "extractor": Placeholders.NO_EXTRACTOR.value,
                }
                for text in prompted_dataset.human_texts
            ]
        )

        dataset = concatenate_datasets([human_dataset, generated_dataset])
        dataset = dataset.shuffle()

        return dataset


class MixCaseGapPacker(MixCasePacker):
    """
    Packer for mixcase task type when using gap-based extractors.
    """

    def __init__(self, config: Config, extractor: Extractor) -> None:
        super().__init__(config=config, extractor=extractor)

    def _build_samples(
        self,
        generations: List[str],
    ) -> Tuple[List[str], List[List[Dict]]]:
        """
        Interleaves generated and human spans to build mixcase samples.
        The `start` and `end` of the labels follow the Python [`start`, `end`)
        convention, i.e., including the `start` element and excluding `end`.
        For instance, given the text: "I like Apolo. I don't like Athenea",
        being the first sentence human-written and the second one generated,
        the labels will be:
        [
            {"start": 0, "end": 14, "label": "human"},
            {"start": 14, "end": 34, "label": "generated"},
        ]

        Args:
            generations (List[str]): list of generations

        Returns:
            Tuple[List[str], List[List[Dict]]]: the interleaved texts and
            the list of labels of each text.
        """
        texts, labels = [], []
        prev_sample = 0
        for idx, sample_boundaries in enumerate(
            self.extractor.workspace["num_boundaries"]
        ):
            # Text w/o sampled boundaries
            if sample_boundaries == 0:
                text = "".join(self.extractor.workspace["human_spans"][idx])
                texts.append(text)
                sample_labels = [
                    LabeledSpan(
                        start=0,
                        end=len(text),
                        label=DetectionLabels.HUMAN.value,
                    ).model_dump()
                ]
            # Text w/ sampled boundaries
            else:
                sample_generations = generations[
                    prev_sample : prev_sample + sample_boundaries
                ]
                # Add a whitespace after generations
                # to be concatenated with the suffix
                sample_generations = [
                    f"{generation} " for generation in sample_generations
                ]

                sample_spans = self.extractor.workspace["human_spans"][idx]
                sample_positions = self.extractor.workspace["positions"][idx]
                sample_labels = []
                added = 0
                prev_label_pos = 0
                # Interleave the generations in the positions determined
                # by the extractor, and computes the labeled spans.
                for i, position in enumerate(sample_positions):
                    # Interleave generation.
                    sample_spans.insert(
                        position + 1 + added, sample_generations[i]
                    )
                    # The generated span starts just after
                    # the prefix until the position `position`.
                    gen_start = len(
                        "".join(sample_spans[: position + 1 + added])
                    )
                    gen_end = gen_start + len(sample_generations[i])
                    # The human span starts from the previous generated span
                    # if exists (prev_label_pos != -1).
                    human_span = LabeledSpan(
                        start=prev_label_pos,
                        end=gen_start,
                        label=DetectionLabels.HUMAN.value,
                    ).model_dump()
                    gen_span = LabeledSpan(
                        start=gen_start,
                        end=gen_end,
                        label=DetectionLabels.GENERATED.value,
                    ).model_dump()
                    sample_labels.append(human_span)
                    sample_labels.append(gen_span)
                    added += 1
                    prev_label_pos = gen_end
                text = "".join(sample_spans)
                # Fill the labels if them do not cover all the text yet.
                # The last span is always human by construction.
                if int(sample_labels[-1]["end"]) < len(text):
                    sample_labels.append(
                        LabeledSpan(
                            start=sample_labels[-1]["end"],
                            end=len(text),
                            label=DetectionLabels.HUMAN.value,
                        ).model_dump()
                    )

                texts.append(text)
            prev_sample += sample_boundaries
            labels.append(sample_labels)
        return texts, labels

    def _pack(self, generations: List[str], **kwargs) -> Dataset:
        """
        Combines and labels the generated and human texts
        when using gap-based extractors (`sentence_gap`, `word_gap`, etc.)

        Args:
            generations (List[str]): list of generated texts.
            kwargs: additional keyword arguments.

        Returns:
            Dataset: a dataset including all the texts.
        """
        prompted_dataset = kwargs.get("prompted_dataset", None)
        if prompted_dataset is None:
            raise DatasetGenerationError(f"prompted_dataset not found: {self}")

        model_name = self.config.model.model_name
        domain = self.config.input.domain
        extractor_name = self.config.input.extractor
        texts, labels = self._build_samples(generations)

        prev_sample = 0
        mixed_samples = []
        for idx, (text, sample_labels) in enumerate(zip(texts, labels)):
            sample_boundaries = self.extractor.workspace["num_boundaries"][idx]
            prompt = prompted_dataset.prompted_texts[
                prev_sample : prev_sample + sample_boundaries
            ] or [Placeholders.NO_PROMPT.value]

            mixed_samples.append(
                {
                    "prompt": prompt,
                    "text": text,
                    "label": sample_labels,
                    "model": model_name,
                    "domain": domain,
                    "extractor": extractor_name,
                }
            )
            prev_sample += self.extractor.workspace["num_boundaries"][idx]

        mixed_dataset = Dataset.from_list(mixed_samples)

        human_dataset = Dataset.from_list(
            [
                {
                    "prompt": [Placeholders.NO_PROMPT.value],
                    "text": text,
                    "label": [
                        LabeledSpan(
                            start=0,
                            end=len(text),
                            label=DetectionLabels.HUMAN.value,
                        ).model_dump()
                    ],
                    "model": DetectionLabels.HUMAN.value,
                    "domain": domain,
                    "extractor": Placeholders.NO_EXTRACTOR.value,
                }
                for text in prompted_dataset.human_texts
            ]
        )

        dataset = concatenate_datasets([human_dataset, mixed_dataset])
        dataset = dataset.shuffle()

        return dataset



import os
from typing import Dict

from ai21 import AI21Client
from ai21.models import ChatMessage

from ..common.logging import get_logger
from ..config import ModelConfig
from .base import TextGenerationModel
from .types import GENERATION_ERROR, CompletionType

_logger = get_logger(__name__)


class AI21Model(TextGenerationModel):
    """
    Generates completions using AI21 models.

    Requires the definition of the `AI21_API_KEY=<api_key>` environment variable.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)
        api_key = os.environ["AI21_API_KEY"]
        num_retries = getattr(self.model_config, "num_retries", 5)
        timeout_sec = getattr(self.model_config, "timeout_sec", 30)
        self.client = AI21Client(
            api_key=api_key, num_retries=num_retries, timeout_sec=timeout_sec
        )

    def generate_completion(
        self,
        prompt: str,
        generation_config: Dict,
    ) -> str:
        completion_fn = (
            self._chat_request
            if self.model_config.api_type == CompletionType.CHAT
            else self._completion_request
        )
        try:
            completion = completion_fn(prompt, generation_config)
        except Exception as e:
            _logger.info(f"Unrecoverable exception during the request: {e}")
            return GENERATION_ERROR
        return completion

    def _chat_request(self, prompt: str, generation_config: Dict) -> str:
        return (
            self.client.chat.create(
                model=self.model_config.model_name,
                messages=[
                    ChatMessage(
                        text=prompt,
                        role="user",
                    ),
                ],
                system="",
                **generation_config,
            )
            .outputs[0]
            .text
        )

    def _completion_request(self, prompt: str, generation_config: Dict) -> str:
        return (
            self.client.completion.create(
                prompt=prompt,
                model=self.model_config.model_name,
                **generation_config,
            )
            .completions[0]
            .data.text
        )



import re
from typing import Dict, List

from datasets import Dataset

from ..config import InputConfig
from ..types import TaskType
from .base import Extractor


class Auxiliary(Extractor):
    """
    Extractor that fills the prompt template with fields from
    a dataset.

    This extractor needs at least one template placeholder, named with
    the name of a field from the dataset, e.g., {summary}.

    This extractor does not need specific arguments.
    """

    def __init__(self, input_config: InputConfig, task_type: TaskType):
        super().__init__(input_config, task_type)

    def _extract(self, dataset: Dataset) -> Dict[str, List[str]]:
        regex = r"\{(\w+)\}"
        columns = re.findall(regex, self.input_config.template)
        return {
            column: dataset[column]
            for column in columns
            if column in dataset.features
        }



from typing import List

from ai21_tokenizer import Tokenizer as AITokenizer

from .base import Tokenizer


class AI21Tokenizer(Tokenizer):
    """
    Tokenizer for AI21 models.

    Requires the definition of the `AI21_API_KEY=<key>` environment variable.
    """

    def __init__(self, model_name: str):
        super().__init__(model_name)
        self.tokenizer = AITokenizer.get_tokenizer()

    def encode(self, text: str) -> List[int]:
        return self.tokenizer.encode(text)

    def decode(self, tokens: List[int]) -> str:
        return self.tokenizer.decode(tokens)



from typing import List

import tiktoken

from .base import Tokenizer


class VertexTokenizer(Tokenizer):
    """
    Tokenizer for VertexAI models.

    VertexAI does not offer tokenizers. GPT-4 Tokenizer is used instead.
    """

    def __init__(self, model_name: str):
        super().__init__(model_name)
        self.tokenizer = tiktoken.encoding_for_model("gpt-4")

    def encode(self, text: str) -> List[int]:
        return self.tokenizer.encode(text)

    def decode(self, tokens: List[int]) -> str:
        return self.tokenizer.decode(tokens)



from typing import Dict, List

from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

from ..config import ModelConfig
from .base import TextGenerationModel
from .types import QUANTIZATION_CONFIGS, CompletionType


class HuggingFaceLocalModel(TextGenerationModel):
    """
    Generates completions using HuggingFace's models locally deployed.
    """

    def __init__(self, model_config: ModelConfig):
        super().__init__(model_config)

        self.model_name = getattr(self.model_config, "model_name")
        self.quantization = getattr(self.model_config, "quantization", "none")
        self.batch_size = getattr(self.model_config, "batch_size", 8)
        self.device = getattr(self.model_config, "device", "cpu")

        self.model = self.__load_model()
        self.tokenizer = self.__load_tokenizer()

    def generate_completion(self, prompt: str, generation_config: Dict) -> str:
        """
        Override `generate_completions` for completeness.
        This method is not used, since generations are done
        with batches using `generate_completions`.
        """
        tokenized = self.tokenizer(
            prompt, truncation=True, padding=True, return_tensors="pt"
        )
        input_ids = tokenized["input_ids"].to(self.model.device)
        attention_mask = tokenized["attention_mask"].to(self.model.device)
        completion = self.model.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            pad_token_id=self.tokenizer.pad_token_id,
            **generation_config,
        )[0]
        return self.tokenizer.decode(
            completion[len(input_ids[0]) :],
            skip_special_tokens=True,
            clean_up_tokenization_spaces=True,
        )

    def generate_completions(
        self,
        prompts: List[str],
        generation_config: Dict,
    ) -> List[str]:
        """
        Overriden method to generate completions using
        HuggingFace's `generate` method with batches
        """
        if self.model_config.api_type == CompletionType.CHAT:
            prompts = [
                self.tokenizer.apply_chat_template(
                    [{"role": "user", "content": prompt}],
                    add_generation_prompt=True,
                    tokenize=False,
                )
                for prompt in prompts
            ]

        tokenized_prompts = self.tokenizer(
            prompts, truncation=True, padding=True, return_tensors="pt"
        )
        completions = []
        for batch_idx in tqdm(
            range(0, len(prompts), self.batch_size),
            desc=f"Generating locally with {self.model_name}",
        ):
            input_ids = tokenized_prompts["input_ids"][
                batch_idx : batch_idx + self.batch_size
            ].to(self.device)
            attention_mask = tokenized_prompts["attention_mask"][
                batch_idx : batch_idx + self.batch_size
            ].to(self.device)
            batch_completions = self.model.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                pad_token_id=self.tokenizer.pad_token_id,
                **generation_config,
            )
            for idx, completion in enumerate(batch_completions):
                completions.append(
                    self.tokenizer.decode(
                        completion[len(input_ids[idx]) :],
                        skip_special_tokens=True,
                        clean_up_tokenization_spaces=True,
                    )
                )
        return completions

    def __load_model(self) -> AutoModelForCausalLM:
        model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
            **QUANTIZATION_CONFIGS[self.quantization],
        )
        if self.quantization == "none":
            model.to(self.device)
        return model

    def __load_tokenizer(self) -> AutoTokenizer:
        tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        if tokenizer.pad_token is None:
            tokenizer.pad_token = tokenizer.eos_token
            tokenizer.pad_token_id = tokenizer.eos_token_id
        tokenizer.padding_side = "left"
        return tokenizer



from typing import List


class TextMachinaError(Exception):
    """
    Base class for TextMachina exceptions.
    """

    ...


class MissingIntegrationError(TextMachinaError):
    """
    Raised when an integration can't be used due to missing dependencies.
    """

    def __init__(self, integration: str):
        self.integration = integration
        msg = (
            f"'{self.integration}' integration can't be used. Please, install"
            f" the corresponding integration:"
            f" `pip install text-machina[{integration}]`."
        )
        super().__init__(msg)


class MissingMetricError(TextMachinaError):
    """
    Raised when a metric can't be used due to missing dependencies.
    """

    def __init__(self, metric: str):
        self.metric = metric
        msg = (
            f"'{self.metric}' metric can't be used. Please, install"
            f" the extra exploration packages"
            f" `pip install text-machina[explore]`."
        )
        super().__init__(msg)


class InvalidInferenceServer(TextMachinaError):
    """
    Raised when an inference server type is not valid
    for the InferenceServerModel.
    """

    def __init__(self, inference_server: str, allowed: List[str]):
        self.inference_server = inference_server
        self.allowed = allowed
        msg = (
            f"Inference server `{self.inference_server}` is not allowed."
            f" Please use one from {', '.join(allowed)}"
        )
        super().__init__(msg)


class InvalidTaskTypeForMetric(TextMachinaError):
    """
    Raised when a metric can't be used due to inadequate task type.
    """

    def __init__(self, metric: str, task_type: str):
        self.metric = metric
        self.task_type = task_type
        msg = f"'{self.metric}' metric cannot be used for task type {self.task_type},"
        super().__init__(msg)


class InvalidProvider(TextMachinaError):
    """
    Raised when a provider is not allowed.
    """

    def __init__(self, provider: str):
        self.provider = provider
        from ..models import MODELS

        msg = (
            f"`provider={provider}` is not allowed. Use one of:"
            f" {', '.join(list(MODELS.keys()))} and install the"
            " corresponding integration."
        )
        super().__init__(msg)


class InvalidMetric(TextMachinaError):
    """
    Raised when a metric is not allowed.
    """

    def __init__(self, metric: str):
        self.metric = metric
        from ..metrics import METRICS

        msg = (
            f"`metric={metric}` is not allowed. Use one of:"
            f" {', '.join(list(METRICS.keys()))} and install"
            " exploration packages."
        )
        super().__init__(msg)


class InvalidExtractor(TextMachinaError):
    """
    Raised when an extractor is not allowed.
    """

    def __init__(self, extractor: str):
        self.extractor = extractor
        from ..extractors import EXTRACTORS

        msg = (
            f"`extractor={self.extractor}` is not allowed. Use one of:"
            f" {', '.join(list(EXTRACTORS.keys()))}."
        )
        super().__init__(msg)


class CombinedEmptyExtractors(TextMachinaError):
    """
    Raised when `combined` extractor is called without a list of extractors.
    """

    def __init__(self):
        msg = (
            "Specify a non-empty list of extractors in the `extractors_list`"
            " field of your config to use the `combined` extractor."
        )
        super().__init__(msg)


class InvalidModelParam(TextMachinaError):
    """
    Raised when a parameter for a text generation model is invalid.
    """

    def __init__(self, msg: str):
        super().__init__(msg)


class DatasetGenerationError(TextMachinaError):
    """
    Raised when an error ocurrs in dataset generators.
    """

    def __init__(self, msg: str):
        super().__init__(msg)


class UnsupportedMetricParam(TextMachinaError):
    """
    Raised when a metric param cannot be used.

    This is not for cases where an incorrect parameter name is provided,
    rather when a parameter is correct but cannot be used.

    See `..metrics.mauve.MAUVEMetric` for an example.
    """

    def __init__(self, param_name: str, metric_name: str):
        self.param_name = param_name
        self.metric_name = metric_name
        msg = f"Parameter {self.param_name} cannot be used for metric {metric_name}"
        super().__init__(msg)


class InvalidLanguage(TextMachinaError):
    """
    Raised when a language is not a valid ISO 639-1 code.
    """

    def __init__(self):
        msg = "`language` must be a valid ISO 639-1 code."
        super().__init__(msg)


class InvalidSpacyModel(TextMachinaError):
    """
    Raised when an Spacy model does not exist for a language.
    """

    def __init__(self, language: str):
        self.language = language
        msg = f"Spacy does not have a model for `language={self.language}`"
        super().__init__(msg)


class ExtractorEmptyColumns(TextMachinaError):
    """
    Raised when an extractor returns empty lists in any field.
    """

    def __init__(self, extractor: str, field: str):
        self.field = field
        msg = f"The extractor {extractor} returned empty '{field}'."
        super().__init__(msg)


class ExtractorInvalidArgs(TextMachinaError):
    """
    Raised when mandatory arguments for an
    extractor are invalid or not provided.
    """

    def __init__(self, extractor: str, extractor_args: List[str]):
        self.extractor = extractor
        self.extractor_args = extractor_args
        msg = (
            f"When using the extractor {extractor}, you must ensure that"
            f" the following arguments: {', '.join(self.extractor_args)}"
            " are valid."
        )
        super().__init__(msg)



from typing import List, Mapping, Type

from ..tokenizers import get_tokenizer
from .length import (
    LengthConstrainer,
    MeanLengthConstrainer,
    MedianLengthConstrainer,
)

LENGTH_CONSTRAINERS: Mapping[str, Type[LengthConstrainer]] = {
    "mean_length": MeanLengthConstrainer,
    "median_length": MedianLengthConstrainer,
}


def get_length_constrainer(
    texts: List[str],
    model_name: str,
    provider: str,
    constrainer_name: str = "mean_length",
    min_tokens: int = 10,
    max_tokens: int = 512,
) -> LengthConstrainer:
    """
    Returns a length constrainer given by the tokenized lengths of the texts
    """
    tokenizer = get_tokenizer(provider, model_name)
    # Clip texts to `max_tokens`
    lengths = list(
        map(lambda x: min(len(tokenizer.encode(x)), max_tokens), texts)
    )
    constrainer = LENGTH_CONSTRAINERS[constrainer_name]
    return constrainer(
        lengths, provider, min_tokens=min_tokens, max_tokens=max_tokens
    )


__all__ = [str(cls) for cls in LENGTH_CONSTRAINERS.values()]
